from math import ceil, floor

import numpy as np
from scipy.interpolate import interp1d

from ghedesigner import VERSION
from ghedesigner.borehole import GHEBorehole
from ghedesigner.borehole_heat_exchangers import get_bhe_object
from ghedesigner.constants import TWO_PI
from ghedesigner.enums import BHPipeType, TimestepType
from ghedesigner.gfunction import GFunction, calc_g_func_for_multiple_lengths
from ghedesigner.ground_loads import HybridLoad
from ghedesigner.media import Grout, Pipe, Soil
from ghedesigner.radial_numerical_borehole import RadialNumericalBH
from ghedesigner.simulation import SimulationParameters
from ghedesigner.utilities import solve_root


class BaseGHE:
    def __init__(
            self,
            v_flow_system: float,
            b_spacing: float,
            bhe_type: BHPipeType,
            fluid,
            borehole: GHEBorehole,
            pipe: Pipe,
            grout: Grout,
            soil: Soil,
            g_function: GFunction,
            sim_params: SimulationParameters,
            hourly_extraction_ground_loads: list,
            field_type="N/A",
            field_specifier="N/A",
    ):

        self.fieldType = field_type
        self.fieldSpecifier = field_specifier
        self.V_flow_system = v_flow_system
        self.B_spacing = b_spacing
        self.nbh = len(g_function.bore_locations)
        self.V_flow_borehole = self.V_flow_system / self.nbh
        m_flow_borehole = self.V_flow_borehole / 1000.0 * fluid.rho
        self.m_flow_borehole = m_flow_borehole

        # Borehole Heat Exchanger
        self.bhe_type = bhe_type
        self.bhe = get_bhe_object(bhe_type, m_flow_borehole, fluid, borehole, pipe, grout, soil)

        # Equivalent borehole Heat Exchanger
        self.bhe_eq = self.bhe.to_single()

        # Radial numerical short time step
        self.radial_numerical = RadialNumericalBH(self.bhe_eq)
        self.radial_numerical.calc_sts_g_functions(self.bhe_eq)

        # gFunction object
        self.gFunction = g_function
        # Additional simulation parameters
        self.sim_params = sim_params
        # Hourly ground extraction loads
        # Building cooling is negative, building heating is positive
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.times = []
        self.loading = None

    def as_dict(self) -> dict:
        output = dict()
        output['title'] = f"GHEDesigner GHE Output - Version {VERSION}"
        output['number_of_boreholes'] = len(self.gFunction.bore_locations)
        output['borehole_depth'] = {'value': self.bhe.b.H, 'units': 'm'}
        output['borehole_spacing'] = {'value': self.B_spacing, 'units': 'm'}
        output['borehole_heat_exchanger'] = self.bhe.as_dict()
        output['equivalent_borehole_heat_exchanger'] = self.bhe_eq.as_dict()
        output['simulation_parameters'] = self.sim_params.as_dict()
        return output

    @staticmethod
    def combine_sts_lts(log_time_lts: list, g_lts: list, log_time_sts: list, g_sts: list) -> interp1d:
        # make sure the short time step doesn't overlap with the long time step
        max_log_time_sts = max(log_time_sts)
        min_log_time_lts = min(log_time_lts)

        if max_log_time_sts < min_log_time_lts:
            log_time = log_time_sts + log_time_lts
            g = g_sts + g_lts
        else:
            # find where to stop in sts
            i = 0
            value = log_time_sts[i]
            while value <= min_log_time_lts:
                i += 1
                value = log_time_sts[i]
            log_time = log_time_sts[0:i] + log_time_lts
            g = g_sts[0:i] + g_lts
        g = interp1d(log_time, g)

        return g

    def grab_g_function(self, b_over_h):
        # interpolate for the Long time step g-function
        g_function, rb_value, _, _ = self.gFunction.g_function_interpolation(b_over_h)
        # correct the long time step for borehole radius
        g_function_corrected = self.gFunction.borehole_radius_correction(
            g_function, rb_value, self.bhe.b.r_b
        )
        # Don't Update the HybridLoad (its dependent on the STS) because
        # it doesn't change the results much, and it slows things down a lot
        # combine the short and long time step g-function
        g = self.combine_sts_lts(
            self.gFunction.log_time,
            g_function_corrected,
            self.radial_numerical.lntts.tolist(),
            self.radial_numerical.g.tolist(),
        )

        g_bhw = self.combine_sts_lts(
            self.gFunction.log_time,
            g_function_corrected,
            self.radial_numerical.lntts.tolist(),
            self.radial_numerical.g_bhw.tolist(),
        )

        return g, g_bhw

    def cost(self, max_eft, min_eft):
        delta_t_max = max_eft - self.sim_params.max_EFT_allowable
        delta_t_min = self.sim_params.min_EFT_allowable - min_eft
        t_excess = max(delta_t_max, delta_t_min)
        return t_excess

    def _simulate_detailed(self, q_dot: np.ndarray, time_values: np.ndarray, g: interp1d):
        # Perform a detailed simulation based on a numpy array of heat rejection
        # rates, Q_dot (Watts) where each load is applied at the time_value
        # (seconds). The g-function can interpolate.
        # Source: Chapter 2 of Advances in Ground Source Heat Pumps

        n = q_dot.size

        # Convert the total load applied to the field to the average over
        # borehole wall rejection rate
        # At time t=0, make the heat rejection rate 0.
        q_dot_b = np.hstack((0.0, q_dot / float(self.nbh)))
        time_values = np.hstack((0.0, time_values))

        q_dot_b_dt = np.hstack((q_dot_b[1:] - q_dot_b[:-1]))

        ts = self.radial_numerical.t_s  # (-)
        two_pi_k = TWO_PI * self.bhe.soil.k  # (W/m.K)
        h = self.bhe.b.H  # (meters)
        tg = self.bhe.soil.ugt  # (Celsius)
        rb = self.bhe.calc_effective_borehole_resistance()  # (m.K/W)
        m_dot = self.bhe.m_flow_borehole  # (kg/s)
        cp = self.bhe.fluid.cp  # (J/kg.s)

        hp_eft = []
        delta_tb = []
        for i in range(1, n + 1):
            # Take the last i elements of the reversed time array
            _time = time_values[i] - time_values[0:i]
            # _time = time_values_reversed[n - i:n]
            g_values = g(np.log((_time * 3600.0) / ts))
            # Tb = Tg + (q_dt * g)  (Equation 2.12)
            delta_tb_i = (q_dot_b_dt[0:i] / h / two_pi_k).dot(g_values)
            # Tf = Tb + q_i * R_b^* (Equation 2.13)
            tb = tg + delta_tb_i
            # Bulk fluid temperature
            tf_bulk = tb + q_dot_b[i] / h * rb
            # T_out = T_f - Q / (2 * m_dot cp)  (Equation 2.14)
            tf_out = tf_bulk - q_dot_b[i] / (2 * m_dot * cp)
            hp_eft.append(tf_out)
            delta_tb.append(delta_tb_i)

        return hp_eft, delta_tb

    def compute_g_functions(self):
        # Compute g-functions for a bracketed solution, based on min and max
        # height
        min_height = self.sim_params.min_height
        max_height = self.sim_params.max_height
        avg_height = (min_height + max_height) / 2.0
        h_values = [min_height, avg_height, max_height]

        coordinates = self.gFunction.bore_locations
        log_time = self.gFunction.log_time

        g_function = calc_g_func_for_multiple_lengths(
            self.B_spacing,
            h_values,
            self.bhe.b.r_b,
            self.bhe.b.D,
            self.bhe.m_flow_borehole,
            self.bhe_type,
            log_time,
            coordinates,
            self.bhe.fluid,
            self.bhe.pipe,
            self.bhe.grout,
            self.bhe.soil,
        )

        self.gFunction = g_function


class GHE(BaseGHE):
    def __init__(
            self,
            v_flow_system: float,
            b_spacing: float,
            bhe_type: BHPipeType,
            fluid,
            borehole: GHEBorehole,
            pipe: Pipe,
            grout: Grout,
            soil: Soil,
            g_function: GFunction,
            sim_params: SimulationParameters,
            hourly_extraction_ground_loads: list,
            field_type="N/A",
            field_specifier="N/A",
            load_years=None,
    ):
        BaseGHE.__init__(
            self,
            v_flow_system,
            b_spacing,
            bhe_type,
            fluid,
            borehole,
            pipe,
            grout,
            soil,
            g_function,
            sim_params,
            hourly_extraction_ground_loads,
            field_type=field_type,
            field_specifier=field_specifier,
        )

        # Split the extraction loads into heating and cooling for input to
        # the HybridLoad object
        if load_years is None:
            load_years = [2019]

        hybrid_load = HybridLoad(
            self.hourly_extraction_ground_loads,
            self.bhe_eq,
            self.radial_numerical,
            sim_params,
            years=load_years)

        # hybrid load object
        self.hybrid_load = hybrid_load

        # List of heat pump exiting fluid temperatures
        self.hp_eft = []
        # list of change in borehole wall temperatures
        self.dTb = []

    def as_dict(self) -> dict:
        output = dict()
        output['base'] = super().as_dict()

        results = dict()
        if len(self.hp_eft) > 0:
            max_hp_eft = max(self.hp_eft)
            min_hp_eft = min(self.hp_eft)
            results['max_hp_entering_temp'] = {'value': max_hp_eft, 'units': 'C'}
            results['min_hp_entering_temp'] = {'value': min_hp_eft, 'units': 'C'}
            t_excess = self.cost(max_hp_eft, min_hp_eft)
            results['excess_fluid_temperature'] = {'value': t_excess, 'units': 'C'}
        results['peak_load_analysis'] = self.hybrid_load.as_dict()

        g_function = dict()
        g_function['coordinates (x[m], y[m])'] = [(x, y) for x, y in self.gFunction.bore_locations]  # TODO: Verify form
        b_over_h = self.B_spacing / self.bhe.b.H
        g, _ = self.grab_g_function(b_over_h)
        total_g_values = g.x.size
        number_lts_g_values = 27
        number_sts_g_values = 50
        sts_step_size = floor((total_g_values - number_lts_g_values) / number_sts_g_values)
        lntts = []
        g_values = []
        for idx in range(0, (total_g_values - number_lts_g_values), sts_step_size):
            lntts.append(g.x[idx].tolist())
            g_values.append(g.y[idx].tolist())
        lntts += g.x[total_g_values - number_lts_g_values: total_g_values].tolist()
        g_values += g.y[total_g_values - number_lts_g_values: total_g_values].tolist()
        pairs = zip(lntts, g_values)
        for lntts_val, g_val in pairs:
            output += f"{lntts_val:0.4f}\t{g_val:0.4f}"
        g_function['lntts, g'] = [*pairs]

        results['g_function_information'] = g_function
        output['simulation_results'] = results

        return output

    def simulate(self, method: TimestepType):
        b = self.B_spacing
        b_over_h = b / self.bhe.b.H

        # Solve for equivalent single U-tube
        self.bhe_eq = self.bhe.to_single()
        # Update short time step object with equivalent single u-tube
        self.radial_numerical.calc_sts_g_functions(self.bhe_eq)
        # Combine the short and long-term g-functions. The long term g-function
        # is interpolated for specific B/H and rb/H values.
        g, _ = self.grab_g_function(b_over_h)

        if method == TimestepType.HYBRID:
            q_dot = self.hybrid_load.load[2:] * 1000.0  # convert to Watts
            time_values = self.hybrid_load.hour[2:]  # convert to seconds
            self.times = time_values
            self.loading = q_dot

            hp_eft, d_tb = self._simulate_detailed(q_dot, time_values, g)
        elif method == TimestepType.HOURLY:
            n_months = self.sim_params.end_month - self.sim_params.start_month + 1
            n_hours = int(n_months / 12.0 * 8760.0)
            q_dot = self.hourly_extraction_ground_loads
            # How many times does q need to be repeated?
            n_years = ceil(n_hours / 8760)
            if len(q_dot) // 8760 < n_years:
                q_dot = q_dot * n_years
            else:
                n_hours = len(q_dot)
            q_dot = -1.0 * np.array(q_dot)  # Convert loads to rejection
            # print("Times:",self.times)
            if len(self.times) == 0:
                self.times = np.arange(1, n_hours + 1, 1)
            t = self.times
            self.loading = q_dot

            hp_eft, d_tb = self._simulate_detailed(q_dot, t, g)
        else:
            raise ValueError("Only hybrid or hourly methods available.")

        self.hp_eft = hp_eft
        self.dTb = d_tb

        return max(hp_eft), min(hp_eft)

    def size(self, method: TimestepType) -> None:
        # Size the ground heat exchanger
        def local_objective(h):
            self.bhe.b.H = h
            max_hp_eft, min_hp_eft = self.simulate(method=method)
            t_excess = self.cost(max_hp_eft, min_hp_eft)
            return t_excess

        # Make the initial guess variable the average of the heights given
        self.bhe.b.H = (self.sim_params.max_height + self.sim_params.min_height) / 2.0
        # bhe.b.H is updated during sizing
        returned_height = solve_root(
            self.bhe.b.H,
            local_objective,
            lower=self.sim_params.min_height,
            upper=self.sim_params.max_height,
            abs_tol=1.0e-6,
            rel_tol=1.0e-6,
            max_iter=50,
        )

        self.bhe.b.H = returned_height



from pygfunction.boreholes import Borehole


class GHEBorehole(Borehole):
    def __init__(self, height, buried_depth, radius, x, y, tilt=0, orientation=0):
        super().__init__(height, buried_depth, radius, x, y, tilt, orientation)

    def to_input(self):
        return {'buried_depth': self.D, 'diameter': self.r_b * 2.0}



from abc import abstractmethod

from ghedesigner.constants import RAD_TO_DEG
from ghedesigner.enums import DesignGeomType


class GeometricConstraints:
    def __init__(self):
        self.type = None

    @abstractmethod
    def to_input(self):
        pass


class GeometricConstraintsNearSquare(GeometricConstraints):
    """
    Geometric constrains for near square design algorithm
    """

    def __init__(self, b: float, length: float):
        super().__init__()
        self.b = b
        self.length = length
        self.type = DesignGeomType.NEARSQUARE

    def to_input(self) -> dict:
        return {'length': self.length,
                'b': self.b,
                'method': DesignGeomType.NEARSQUARE.name}


class GeometricConstraintsRectangle(GeometricConstraints):
    """
    Geometric constraints for rectangular design algorithm
    """

    def __init__(self, width: float, length: float, b_min: float, b_max_x: float):
        super().__init__()
        self.width = width
        self.length = length
        self.b_min = b_min
        self.b_max_x = b_max_x
        self.type = DesignGeomType.RECTANGLE

    def to_input(self) -> dict:
        return {'length': self.length,
                'width': self.width,
                'b_min': self.b_min,
                'b_max': self.b_max_x,
                'method': DesignGeomType.RECTANGLE.name}


class GeometricConstraintsBiRectangle(GeometricConstraints):
    """
    Geometric constraints for bi-rectangle design algorithm
    """

    def __init__(self, width: float, length: float, b_min: float, b_max_x: float, b_max_y: float):
        super().__init__()
        self.width = width
        self.length = length
        self.b_min = b_min
        self.b_max_x = b_max_x
        self.b_max_y = b_max_y
        self.type = DesignGeomType.BIRECTANGLE

    def to_input(self) -> dict:
        return {'length': self.length,
                'width': self.width,
                'b_min': self.b_min,
                'b_max_x': self.b_max_x,
                'b_max_y': self.b_max_y,
                'method': DesignGeomType.BIRECTANGLE.name}


class GeometricConstraintsBiRectangleConstrained(GeometricConstraints):
    """
    Geometric constraints for bi-rectangle constrained design algorithm
    """

    def __init__(self, b_min: float, b_max_x: float, b_max_y: float, property_boundary, no_go_boundaries):
        super().__init__()
        self.b_min = b_min
        self.b_max_x = b_max_x
        self.b_max_y = b_max_y

        if len(no_go_boundaries) > 0 and isinstance(no_go_boundaries[0][0], (int, float)):
            self.no_go_boundaries = [no_go_boundaries]
        else:
            self.no_go_boundaries = no_go_boundaries

        if len(property_boundary) > 0 and isinstance(property_boundary[0][0], (int, float)):
            self.property_boundary = [property_boundary]
        else:
            self.property_boundary = property_boundary

        self.type = DesignGeomType.BIRECTANGLECONSTRAINED

    def to_input(self) -> dict:
        return {'b_min': self.b_min,
                'b_max_x': self.b_max_x,
                'b_max_y': self.b_max_y,
                'property_boundary': self.property_boundary,
                'no_go_boundaries': self.no_go_boundaries,
                'method': DesignGeomType.BIRECTANGLECONSTRAINED.name}


class GeometricConstraintsBiZoned(GeometricConstraintsBiRectangle):
    """
    Geometric constraints for bi-zoned design algorithm
    """

    def __init__(self, width: float, length: float, b_min: float, b_max_x: float, b_max_y: float):
        super().__init__(width, length, b_min, b_max_x, b_max_y)
        self.type = DesignGeomType.BIZONEDRECTANGLE

    def to_input(self) -> dict:
        return {'length': self.length,
                'width': self.width,
                'b_min': self.b_min,
                'b_max_x': self.b_max_x,
                'b_max_y': self.b_max_y,
                'method': DesignGeomType.BIZONEDRECTANGLE.name}


class GeometricConstraintsRowWise(GeometricConstraints):
    """
    Geometric constraints for rowwise design algorithm
    """

    def __init__(self,
                 perimeter_spacing_ratio: float,
                 min_spacing: float,
                 max_spacing: float,
                 spacing_step: float,
                 min_rotation: float,
                 max_rotation: float,
                 rotate_step: float,
                 property_boundary,
                 no_go_boundaries):
        super().__init__()
        self.perimeter_spacing_ratio = perimeter_spacing_ratio
        self.min_spacing = min_spacing
        self.max_spacing = max_spacing
        self.spacing_step = spacing_step
        self.min_rotation = min_rotation
        self.max_rotation = max_rotation
        self.rotate_step = rotate_step
        self.property_boundary = property_boundary
        self.no_go_boundaries = no_go_boundaries
        self.type = DesignGeomType.ROWWISE

    def to_input(self) -> dict:
        return {'perimeter_spacing_ratio': self.perimeter_spacing_ratio,
                'min_spacing': self.min_spacing,
                'max_spacing': self.max_spacing,
                'spacing_step': self.spacing_step,
                'min_rotation': self.min_rotation * RAD_TO_DEG,
                'max_rotation': self.max_rotation * RAD_TO_DEG,
                'rotate_step': self.rotate_step,
                'property_boundary': self.property_boundary,
                'no_go_boundaries': self.no_go_boundaries,
                'method': DesignGeomType.ROWWISE.name}



from enum import Enum, auto


class BHPipeType(Enum):
    COAXIAL = auto()
    DOUBLEUTUBEPARALLEL = auto()
    DOUBLEUTUBESERIES = auto()
    SINGLEUTUBE = auto()


class DoubleUTubeConnType(Enum):
    PARALLEL = auto()
    SERIES = auto()


class TimestepType(Enum):
    HOURLY = auto()
    HYBRID = auto()


class DesignGeomType(Enum):
    BIRECTANGLE = auto()
    BIRECTANGLECONSTRAINED = auto()
    BIZONEDRECTANGLE = auto()
    NEARSQUARE = auto()
    RECTANGLE = auto()
    ROWWISE = auto()


class FlowConfigType(Enum):
    BOREHOLE = auto()
    SYSTEM = auto()


class FluidType(Enum):
    ETHYLALCOHOL = auto()
    ETHYLENEGLYCOL = auto()
    METHYLALCOHOL = auto()
    PROPYLENEGLYCOL = auto()
    WATER = auto()



from json import loads, dumps
from pathlib import Path
from sys import exit, stderr
from time import time
from typing import List, Optional, Union

import click
from jsonschema import ValidationError

from ghedesigner import VERSION
from ghedesigner.borehole import GHEBorehole
from ghedesigner.constants import DEG_TO_RAD
from ghedesigner.design import AnyBisectionType, DesignBase, DesignNearSquare, DesignRectangle, DesignBiRectangle
from ghedesigner.design import DesignBiZoned, DesignBiRectangleConstrained, DesignRowWise
from ghedesigner.enums import BHPipeType, TimestepType, DesignGeomType, FlowConfigType
from ghedesigner.geometry import GeometricConstraints, GeometricConstraintsRectangle, GeometricConstraintsNearSquare
from ghedesigner.geometry import GeometricConstraintsBiRectangle, GeometricConstraintsBiZoned
from ghedesigner.geometry import GeometricConstraintsBiRectangleConstrained, GeometricConstraintsRowWise
from ghedesigner.media import GHEFluid, Grout, Pipe, Soil
from ghedesigner.output import OutputManager
from ghedesigner.simulation import SimulationParameters
from ghedesigner.utilities import write_idf
from ghedesigner.validate import validate_input_file


class GHEManager:

    def __init__(self):
        self._fluid: Optional[GHEFluid] = None
        self._grout: Optional[Grout] = None
        self._soil: Optional[Soil] = None
        self._pipe: Optional[Pipe] = None
        self.pipe_type: Optional[BHPipeType] = None
        self._borehole: Optional[GHEBorehole] = None
        self._simulation_parameters: Optional[SimulationParameters] = None
        self._ground_loads: Optional[List[float]] = None
        # OK so geometric_constraints is tricky.  We have base classes, yay!
        # Unfortunately, the functionality between the child classes is not actually
        # collapsed into a base class function ... yet.  So there will be complaints
        # about types temporarily.  It's going in the right direction though.
        self.geom_type: Optional[DesignGeomType] = None
        self._geometric_constraints: Optional[GeometricConstraints] = None
        self._design: Optional[DesignBase] = None
        self._search: Optional[AnyBisectionType] = None
        self.results: Optional[OutputManager] = None

        # some things for results
        self._search_time: int = 0
        self.summary_results: dict = {}

    def set_design_geometry_type(self, design_geometry_str: str, throw: bool = True) -> int:
        """
        Sets the design type.

        :param design_geometry_str: design geometry input string.
        :param throw: By default, function will raise an exception on error, override to false to not raise exception
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        design_geometry_str = str(design_geometry_str).upper()
        if design_geometry_str == DesignGeomType.BIRECTANGLE.name:
            self.geom_type = DesignGeomType.BIRECTANGLE
        elif design_geometry_str == DesignGeomType.BIRECTANGLECONSTRAINED.name:
            self.geom_type = DesignGeomType.BIRECTANGLECONSTRAINED
        elif design_geometry_str == DesignGeomType.BIZONEDRECTANGLE.name:
            self.geom_type = DesignGeomType.BIZONEDRECTANGLE
        elif design_geometry_str == DesignGeomType.NEARSQUARE.name:
            self.geom_type = DesignGeomType.NEARSQUARE
        elif design_geometry_str == DesignGeomType.RECTANGLE.name:
            self.geom_type = DesignGeomType.RECTANGLE
        elif design_geometry_str == DesignGeomType.ROWWISE.name:
            self.geom_type = DesignGeomType.ROWWISE
        else:
            message = "Geometry constraint method not supported."
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1

        return 0

    def set_pipe_type(self, bh_pipe_str: str, throw: bool = True) -> int:
        """
        Sets the borehole pipe type.

        :param bh_pipe_str: pipe type input string.
        :param throw: By default, function will raise an exception on error, override to false to not raise exception
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        bh_pipe_str = str(bh_pipe_str).upper()
        if bh_pipe_str == BHPipeType.SINGLEUTUBE.name:
            self.pipe_type = BHPipeType.SINGLEUTUBE
        elif bh_pipe_str == BHPipeType.DOUBLEUTUBEPARALLEL.name:
            self.pipe_type = BHPipeType.DOUBLEUTUBEPARALLEL
        elif bh_pipe_str == BHPipeType.DOUBLEUTUBESERIES.name:
            self.pipe_type = BHPipeType.DOUBLEUTUBESERIES
        elif bh_pipe_str == BHPipeType.COAXIAL.name:
            self.pipe_type = BHPipeType.COAXIAL
        else:
            message = f"Borehole pipe type \"{bh_pipe_str}\" not supported."
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1

        return 0

    def set_fluid(self, fluid_name: str = "Water", concentration_percent: float = 0.0, temperature: float = 20.0,
                  throw: bool = True) -> int:
        """
        Sets the fluid instance.

        :param fluid_name: fluid name input string.
        :param concentration_percent: concentration percent of antifreeze mixture.
        :param temperature: design fluid temperature, in C.
        :param throw: By default, function will raise an exception on error, override to false to not raise exception
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        try:
            self._fluid = GHEFluid(fluid_str=fluid_name,
                                   percent=concentration_percent,
                                   temperature=temperature)
            return 0
        except ValueError:
            message = "Invalid fluid property input data."
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1

    def set_grout(self, conductivity: float, rho_cp: float) -> int:
        """
        Sets the grout instance.

        :param conductivity: thermal conductivity, in W/m-K.
        :param rho_cp: volumetric heat capacity, in J/m^3-K.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        self._grout = Grout(conductivity, rho_cp)
        return 0

    def set_soil(self, conductivity: float, rho_cp: float, undisturbed_temp: float) -> int:
        """
        Sets the soil instance.

        :param conductivity: thermal conductivity, in W/m-K.
        :param rho_cp: volumetric heat capacity, in J/m^3-K.
        :param undisturbed_temp: undisturbed soil temperature, in C.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        self._soil = Soil(conductivity, rho_cp, undisturbed_temp)
        return 0

    def set_single_u_tube_pipe(self, inner_diameter: float, outer_diameter: float, shank_spacing: float,
                               roughness: float, conductivity: float, rho_cp: float) -> int:
        """
        Sets the pipe instance for a single u-tube pipe.

        :param inner_diameter: inner pipe diameter, in m.
        :param outer_diameter: outer pipe diameter, in m.
        :param shank_spacing: shank spacing between the u-tube legs, in m, as measured edge-to-edge.
        :param roughness: pipe surface roughness, in m.
        :param conductivity: thermal conductivity, in W/m-K.
        :param rho_cp: volumetric heat capacity, in J/m^3-K.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """

        r_in = inner_diameter / 2.0
        r_out = outer_diameter / 2.0

        self.pipe_type = BHPipeType.SINGLEUTUBE
        pipe_positions = Pipe.place_pipes(shank_spacing, r_out, 1)
        self._pipe = Pipe(pipe_positions, r_in, r_out, shank_spacing, roughness, conductivity, rho_cp)
        return 0

    def set_double_u_tube_pipe_parallel(self, inner_diameter: float, outer_diameter: float, shank_spacing: float,
                                        roughness: float, conductivity: float, rho_cp: float) -> int:
        """
        Sets the pipe instance for a double u-tube pipe in a parallel configuration.

        :param inner_diameter: inner pipe diameter, in m.
        :param outer_diameter: outer pipe diameter, in m.
        :param shank_spacing: shank spacing between the u-tube legs, in m, as measured edge-to-edge.
        :param roughness: pipe surface roughness, in m.
        :param conductivity: thermal conductivity, in W/m-K.
        :param rho_cp: volumetric heat capacity, in J/m^3-K.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """

        r_in = inner_diameter / 2.0
        r_out = outer_diameter / 2.0

        self.pipe_type = BHPipeType.DOUBLEUTUBEPARALLEL
        pipe_positions = Pipe.place_pipes(shank_spacing, r_out, 2)
        self._pipe = Pipe(pipe_positions, r_in, r_out, shank_spacing, roughness, conductivity, rho_cp)
        return 0

    def set_double_u_tube_pipe_series(self, inner_diameter: float, outer_diameter: float, shank_spacing: float,
                                      roughness: float, conductivity: float, rho_cp: float) -> int:
        """
        Sets the pipe instance for a double u-tube pipe in a series configuration.

        :param inner_diameter: inner pipe diameter, in m.
        :param outer_diameter: outer pipe diameter, in m.
        :param shank_spacing: shank spacing between the u-tube legs, in m, as measured edge-to-edge.
        :param roughness: pipe surface roughness, in m.
        :param conductivity: thermal conductivity, in W/m-K.
        :param rho_cp: volumetric heat capacity, in J/m^3-K.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """

        r_in = inner_diameter / 2.0
        r_out = outer_diameter / 2.0

        self.pipe_type = BHPipeType.DOUBLEUTUBESERIES
        pipe_positions = Pipe.place_pipes(shank_spacing, r_out, 2)
        self._pipe = Pipe(pipe_positions, r_in, r_out, shank_spacing, roughness, conductivity, rho_cp)
        return 0

    def set_coaxial_pipe(self, inner_pipe_d_in: float, inner_pipe_d_out: float, outer_pipe_d_in: float,
                         outer_pipe_d_out: float, roughness: float, conductivity_inner: float,
                         conductivity_outer: float,
                         rho_cp: float) -> int:
        """
        Sets the pipe instance for a coaxial pipe.

        :param inner_pipe_d_in: inner pipe inner diameter, in m.
        :param inner_pipe_d_out: inner pipe outer diameter, in m.
        :param outer_pipe_d_in: outer pipe inner diameter, in m.
        :param outer_pipe_d_out: outer pipe outer diameter, in m.
        :param roughness: pipe surface roughness, in m.
        :param conductivity_inner: thermal conductivity of inner pipe, in W/m-K.
        :param conductivity_outer: thermal conductivity of outer pipe, in W/m-K.
        :param rho_cp: volumetric heat capacity, in J/m^3-K.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """

        self.pipe_type = BHPipeType.COAXIAL

        # Note: This convention is different from pygfunction
        r_inner = [inner_pipe_d_in / 2.0, inner_pipe_d_out / 2.0]  # The radii of the inner pipe from in to out
        r_outer = [outer_pipe_d_in / 2.0, outer_pipe_d_out / 2.0]  # The radii of the outer pipe from in to out
        k_p = [conductivity_inner, conductivity_outer]
        self._pipe = Pipe((0, 0), r_inner, r_outer, 0, roughness, k_p, rho_cp)
        return 0

    def set_borehole(self, height: float, buried_depth: float, diameter: float) -> int:
        """
        Sets the borehole instance

        :param height: height, or active length, of the borehole, in m.
        :param buried_depth: depth of top of borehole below the ground surface, in m.
        :param diameter: diameter of the borehole, in m.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        radius = diameter / 2.0
        self._borehole = GHEBorehole(height, buried_depth, radius, x=0.0, y=0.0)
        return 0

    def set_simulation_parameters(
            self, num_months: int, max_eft: float, min_eft: float, max_height: float, min_height: float,
            max_boreholes: int = None, continue_if_design_unmet: bool = False
    ) -> int:
        """
        Sets the simulation parameters

        :param num_months: number of months in simulation.
        :param max_eft: maximum heat pump entering fluid temperature, in C.
        :param min_eft: minimum heat pump entering fluid temperature, in C.
        :param max_height: maximum height of borehole, in m.
        :param min_height: minimum height of borehole, in m.
        :param max_boreholes: maximum boreholes in search algorithms.
        :param continue_if_design_unmet: continues to process if design unmet.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        self._simulation_parameters = SimulationParameters(
            1,
            num_months,
            max_eft,
            min_eft,
            max_height,
            min_height,
            max_boreholes,
            continue_if_design_unmet
        )
        return 0

    def set_ground_loads_from_hourly_list(self, hourly_ground_loads: List[float]) -> int:
        """
        Sets the ground loads based on a list input.

        :param hourly_ground_loads: annual, hourly ground loads, in W.
         positive values indicate heat extraction, negative values indicate heat rejection.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        # TODO: Add API methods for different load inputs
        self._ground_loads = hourly_ground_loads
        return 0

    def set_geometry_constraints_near_square(self, b: float, length: float) -> int:
        """
        Sets the geometry constraints for the near-square design method.

        :param b: borehole-to-borehole spacing, in m.
        :param length: side length of the sizing domain, in m.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        self._geometric_constraints = GeometricConstraintsNearSquare(b, length)
        return 0

    def set_geometry_constraints_rectangle(self, length: float, width: float, b_min: float, b_max: float) -> int:
        """
        Sets the geometry constraints for the rectangle design method.

        :param length: side length of the sizing domain, in m.
        :param width: side width of the sizing domain, in m.
        :param b_min: minimum borehole-to-borehole spacing, in m.
        :param b_max: maximum borehole-to-borehole spacing, in m.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        self.geom_type = DesignGeomType.RECTANGLE
        self._geometric_constraints = GeometricConstraintsRectangle(width, length, b_min, b_max)
        return 0

    def set_geometry_constraints_bi_rectangle(self, length: float, width: float, b_min: float,
                                              b_max_x: float, b_max_y: float) -> int:
        """
        Sets the geometry constraints for the bi-rectangle design method.

        :param length: side length of the sizing domain, in m.
        :param width: side width of the sizing domain, in m.
        :param b_min: minimum borehole-to-borehole spacing, in m.
        :param b_max_x: maximum borehole-to-borehole spacing in the x-direction, in m.
        :param b_max_y: maximum borehole-to-borehole spacing in the y-direction, in m.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        self.geom_type = DesignGeomType.BIRECTANGLE
        self._geometric_constraints = GeometricConstraintsBiRectangle(width, length, b_min, b_max_x, b_max_y)
        return 0

    def set_geometry_constraints_bi_zoned_rectangle(self, length: float, width: float, b_min: float,
                                                    b_max_x: float, b_max_y: float) -> int:
        """
        Sets the geometry constraints for the bi-zoned rectangle design method.

        :param length: side length of the sizing domain, in m.
        :param width: side width of the sizing domain, in m.
        :param b_min: minimum borehole-to-borehole spacing, in m.
        :param b_max_x: maximum borehole-to-borehole spacing in the x-direction, in m.
        :param b_max_y: maximum borehole-to-borehole spacing in the y-direction, in m.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        self.geom_type = DesignGeomType.BIZONEDRECTANGLE
        self._geometric_constraints = GeometricConstraintsBiZoned(width, length, b_min, b_max_x, b_max_y)
        return 0

    def set_geometry_constraints_bi_rectangle_constrained(self, b_min: float, b_max_x: float, b_max_y: float,
                                                          property_boundary: list, no_go_boundaries: list) -> int:
        """
        Sets the geometry constraints for the bi-rectangle constrained design method.

        :param b_min: minimum borehole-to-borehole spacing, in m.
        :param b_max_x: maximum borehole-to-borehole spacing in the x-direction, in m.
        :param b_max_y: maximum borehole-to-borehole spacing in the y-direction, in m.
        :param property_boundary: property boundary points, in m.
        :param no_go_boundaries: boundary points for no-go zones, in m.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """
        self.geom_type = DesignGeomType.BIRECTANGLECONSTRAINED
        self._geometric_constraints = GeometricConstraintsBiRectangleConstrained(b_min, b_max_x, b_max_y,
                                                                                 property_boundary, no_go_boundaries)
        return 0

    def set_geometry_constraints_rowwise(self, perimeter_spacing_ratio: Union[float, None],
                                         max_spacing: float, min_spacing: float, spacing_step: float,
                                         max_rotation: float, min_rotation: float, rotate_step: float,
                                         property_boundary: list, no_go_boundaries: list) -> int:
        """
        Sets the geometry constraints for the row-wise design method.

        :param perimeter_spacing_ratio: the ratio between the minimum spacing between
            boreholes placed along the property and no-go zones and the standard borehole-to-borehole
            spacing used for internal boreholes.
        :param max_spacing: the largest minimum spacing that will be used to generate a RowWise field.
        :param min_spacing: the smallest minimum spacing that will be used to generate a RowWise field.
        :param spacing_step: the distance in spacing from the design found in the first part of first
            search to exhaustively check in the second part.
        :param max_rotation: the maximum rotation of the rows of each field relative to horizontal that
            will be used in the search.
        :param min_rotation: the minimum rotation of the rows of each field relative to horizontal that
            will be used in the search.
        :param rotate_step: step size for field rotation search.
        :param property_boundary: property boundary points.
        :param no_go_boundaries: boundary points for no-go zones.
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """

        # convert from degrees to radians
        max_rotation = max_rotation * DEG_TO_RAD
        min_rotation = min_rotation * DEG_TO_RAD

        self.geom_type = DesignGeomType.ROWWISE
        self._geometric_constraints = GeometricConstraintsRowWise(perimeter_spacing_ratio,
                                                                  min_spacing, max_spacing, spacing_step,
                                                                  min_rotation, max_rotation, rotate_step,
                                                                  property_boundary, no_go_boundaries)
        return 0

    def set_design(self, flow_rate: float, flow_type_str: str, throw: bool = True) -> int:
        """
        Set the design method.

        :param flow_rate: design flow rate, in lps.
        :param flow_type_str: flow type string input.
        :param throw: By default, function will raise an exception on error, override to false to not raise exception
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """

        flow_type_str = flow_type_str.upper()
        if flow_type_str == FlowConfigType.SYSTEM.name:
            flow_type = FlowConfigType.SYSTEM
        elif flow_type_str == FlowConfigType.BOREHOLE.name:
            flow_type = FlowConfigType.BOREHOLE
        else:
            message = f"FlowConfig \"{flow_type_str}\" is not implemented."
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1

        if self._geometric_constraints.type is None:
            message = "Geometric constraints must be set before `set_design` is called."
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1

        if self._geometric_constraints.type == DesignGeomType.NEARSQUARE:
            # temporary disable of the type checker because of the _geometric_constraints member
            # noinspection PyTypeChecker
            self._design = DesignNearSquare(
                flow_rate,
                self._borehole,
                self.pipe_type,
                self._fluid,
                self._pipe,
                self._grout,
                self._soil,
                self._simulation_parameters,
                self._geometric_constraints,
                self._ground_loads,
                flow_type=flow_type,
                method=TimestepType.HYBRID,
            )
        elif self._geometric_constraints.type == DesignGeomType.RECTANGLE:
            # temporary disable of the type checker because of the _geometric_constraints member
            # noinspection PyTypeChecker
            self._design = DesignRectangle(
                flow_rate,
                self._borehole,
                self.pipe_type,
                self._fluid,
                self._pipe,
                self._grout,
                self._soil,
                self._simulation_parameters,
                self._geometric_constraints,
                self._ground_loads,
                flow_type=flow_type,
                method=TimestepType.HYBRID,
            )
        elif self._geometric_constraints.type == DesignGeomType.BIRECTANGLE:
            # temporary disable of the type checker because of the _geometric_constraints member
            # noinspection PyTypeChecker
            self._design = DesignBiRectangle(
                flow_rate,
                self._borehole,
                self.pipe_type,
                self._fluid,
                self._pipe,
                self._grout,
                self._soil,
                self._simulation_parameters,
                self._geometric_constraints,
                self._ground_loads,
                flow_type=flow_type,
                method=TimestepType.HYBRID,
            )
        elif self._geometric_constraints.type == DesignGeomType.BIZONEDRECTANGLE:
            # temporary disable of the type checker because of the _geometric_constraints member
            # noinspection PyTypeChecker
            self._design = DesignBiZoned(
                flow_rate,
                self._borehole,
                self.pipe_type,
                self._fluid,
                self._pipe,
                self._grout,
                self._soil,
                self._simulation_parameters,
                self._geometric_constraints,
                self._ground_loads,
                flow_type=flow_type,
                method=TimestepType.HYBRID,
            )
        elif self._geometric_constraints.type == DesignGeomType.BIRECTANGLECONSTRAINED:
            # temporary disable of the type checker because of the _geometric_constraints member
            # noinspection PyTypeChecker
            self._design = DesignBiRectangleConstrained(
                flow_rate,
                self._borehole,
                self.pipe_type,
                self._fluid,
                self._pipe,
                self._grout,
                self._soil,
                self._simulation_parameters,
                self._geometric_constraints,
                self._ground_loads,
                flow_type=flow_type,
                method=TimestepType.HYBRID,
            )
        elif self._geometric_constraints.type == DesignGeomType.ROWWISE:
            # temporary disable of the type checker because of the _geometric_constraints member
            # noinspection PyTypeChecker
            self._design = DesignRowWise(
                flow_rate,
                self._borehole,
                self.pipe_type,
                self._fluid,
                self._pipe,
                self._grout,
                self._soil,
                self._simulation_parameters,
                self._geometric_constraints,
                self._ground_loads,
                flow_type=flow_type,
                method=TimestepType.HYBRID,
            )
        else:
            message = "This design method has not been implemented"
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1
        return 0

    def find_design(self, throw: bool = True) -> int:
        """
        Calls design methods to execute sizing.

        :param throw: By default, function will raise an exception on error, override to false to not raise exception
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """

        if any([x is None for x in [
            self._fluid,
            self._grout,
            self._soil,
            self._pipe,
            self._borehole,
            self._simulation_parameters,
            self._ground_loads,
            self._geometric_constraints,
            self._design,
        ]]):
            message = "All GHE properties must be set before GHEManager.find_design is called."
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1

        start_time = time()
        self._search = self._design.find_design()
        self._search.ghe.compute_g_functions()
        self._search_time = time() - start_time
        self._search.ghe.size(method=TimestepType.HYBRID)
        return 0

    def prepare_results(self, project_name: str, note: str, author: str, iteration_name: str):
        """
        Prepares the output results.
        """
        self.results = OutputManager(
            self._search,
            self._search_time,
            project_name,
            note,
            author,
            iteration_name,
            load_method=TimestepType.HYBRID,
        )

    def write_output_files(self, output_directory: Path, output_file_suffix: str = ""):
        """
        Writes the output files.

        :param output_directory: output directory for output files.
        :param output_file_suffix: adds a string suffix to the output files.
        """
        self.results.write_all_output_files(output_directory=output_directory, file_suffix=output_file_suffix)

    def write_input_file(self, output_file_path: Path, throw: bool = True) -> int:
        """
        Writes an input file based on current simulation configuration.

        :param output_file_path: output directory to write input file.
        :param throw: By default, function will raise an exception on error, override to false to not raise exception
        :returns: Zero if successful, nonzero if failure
        :rtype: int
        """

        # TODO: geometric constraints are currently held in two places
        #       SimulationParameters and GeometricConstraints
        #       these should be consolidated
        d_geo = self._geometric_constraints.to_input()
        d_geo['max_height'] = self._simulation_parameters.max_height
        d_geo['min_height'] = self._simulation_parameters.min_height

        # TODO: data held in different places
        d_des = self._design.to_input()
        d_des['max_eft'] = self._simulation_parameters.max_EFT_allowable
        d_des['min_eft'] = self._simulation_parameters.min_EFT_allowable

        if self._simulation_parameters.max_boreholes is not None:
            d_des['max_boreholes'] = self._simulation_parameters.max_boreholes
        if self._simulation_parameters.continue_if_design_unmet is True:
            d_des['continue_if_design_unmet'] = self._simulation_parameters.continue_if_design_unmet

        # pipe data
        d_pipe = {'rho_cp': self._pipe.rhoCp, 'roughness': self._pipe.roughness}

        if self.pipe_type in [BHPipeType.SINGLEUTUBE, BHPipeType.DOUBLEUTUBEPARALLEL, BHPipeType.DOUBLEUTUBESERIES]:
            d_pipe['inner_diameter'] = self._pipe.r_in * 2.0
            d_pipe['outer_diameter'] = self._pipe.r_out * 2.0
            d_pipe['shank_spacing'] = self._pipe.s
            d_pipe['conductivity'] = self._pipe.k
        elif self.pipe_type == BHPipeType.COAXIAL:
            d_pipe['inner_pipe_d_in'] = self._pipe.r_in[0] * 2.0
            d_pipe['inner_pipe_d_out'] = self._pipe.r_in[1] * 2.0
            d_pipe['outer_pipe_d_in'] = self._pipe.r_out[0] * 2.0
            d_pipe['outer_pipe_d_out'] = self._pipe.r_out[1] * 2.0
            d_pipe['conductivity_inner'] = self._pipe.k[0]
            d_pipe['conductivity_outer'] = self._pipe.k[1]
        else:
            message = 'Invalid pipe type'
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1

        if self.pipe_type == BHPipeType.SINGLEUTUBE:
            d_pipe['arrangement'] = BHPipeType.SINGLEUTUBE.name
        elif self.pipe_type == BHPipeType.DOUBLEUTUBEPARALLEL:
            d_pipe['arrangement'] = BHPipeType.DOUBLEUTUBEPARALLEL.name
        elif self.pipe_type == BHPipeType.DOUBLEUTUBESERIES:
            d_pipe['arrangement'] = BHPipeType.DOUBLEUTUBESERIES.name
        elif self.pipe_type == BHPipeType.COAXIAL:
            d_pipe['arrangement'] = BHPipeType.COAXIAL.name
        else:
            message = 'Invalid pipe type'
            print(message, file=stderr)
            if throw:
                raise ValueError(message)
            return 1

        d = {
            'version': VERSION,
            'fluid': self._fluid.to_input(),
            'grout': self._grout.to_input(),
            'soil': self._soil.to_input(),
            'pipe': d_pipe,
            'borehole': self._borehole.to_input(),
            'simulation': self._simulation_parameters.to_input(),
            'geometric_constraints': d_geo,
            'design': d_des,
            'loads': {'ground_loads': list(self._ground_loads)}
        }

        with open(output_file_path, 'w') as f:
            f.write(dumps(d, sort_keys=True, indent=2, separators=(',', ': ')))
        return 0


def run_manager_from_cli_worker(input_file_path: Path, output_directory: Path) -> int:
    """
    Worker function to run simulation.

    :param input_file_path: path to input file.
    :param output_directory: path to write output files.
    """

    if not input_file_path.exists():
        print(f"No input file found at {input_file_path}, aborting", file=stderr)
        return 1

    if output_directory is None:
        print("Output directory must be passed as an argument, aborting", file=stderr)
        return 1

    # validate inputs against schema before doing anything
    if validate_input_file(input_file_path) != 0:
        return 1

    inputs = loads(input_file_path.read_text())

    ghe = GHEManager()

    version = inputs['version']

    if version != VERSION:
        print("Mismatched version, could be a problem", file=stderr)

    fluid_props = inputs['fluid']  # type: dict
    grout_props = inputs['grout']  # type: dict
    soil_props = inputs['soil']  # type: dict
    pipe_props = inputs['pipe']  # type: dict
    borehole_props = inputs['borehole']  # type: dict
    sim_props = inputs['simulation']  # type: dict
    constraint_props = inputs['geometric_constraints']  # type: dict
    design_props = inputs['design']  # type: dict
    ground_load_props = inputs['loads']['ground_loads']  # type: list

    ghe.set_fluid(**fluid_props, throw=False)
    ghe.set_grout(**grout_props)
    ghe.set_soil(**soil_props)
    ghe.set_pipe_type(pipe_props["arrangement"], throw=False)

    if ghe.pipe_type == BHPipeType.SINGLEUTUBE:
        ghe.set_single_u_tube_pipe(
            inner_diameter=pipe_props["inner_diameter"],
            outer_diameter=pipe_props["outer_diameter"],
            shank_spacing=pipe_props["shank_spacing"],
            roughness=pipe_props["roughness"],
            conductivity=pipe_props["conductivity"],
            rho_cp=pipe_props["rho_cp"]
        )
    elif ghe.pipe_type == BHPipeType.DOUBLEUTUBEPARALLEL:
        ghe.set_double_u_tube_pipe_parallel(
            inner_diameter=pipe_props["inner_diameter"],
            outer_diameter=pipe_props["outer_diameter"],
            shank_spacing=pipe_props["shank_spacing"],
            roughness=pipe_props["roughness"],
            conductivity=pipe_props["conductivity"],
            rho_cp=pipe_props["rho_cp"]
        )
    elif ghe.pipe_type == BHPipeType.DOUBLEUTUBESERIES:
        ghe.set_double_u_tube_pipe_series(
            inner_diameter=pipe_props["inner_diameter"],
            outer_diameter=pipe_props["outer_diameter"],
            shank_spacing=pipe_props["shank_spacing"],
            roughness=pipe_props["roughness"],
            conductivity=pipe_props["conductivity"],
            rho_cp=pipe_props["rho_cp"]
        )
    elif ghe.pipe_type == BHPipeType.COAXIAL:
        ghe.set_coaxial_pipe(
            inner_pipe_d_in=pipe_props["inner_pipe_d_in"],
            inner_pipe_d_out=pipe_props["inner_pipe_d_out"],
            outer_pipe_d_in=pipe_props["outer_pipe_d_in"],
            outer_pipe_d_out=pipe_props["outer_pipe_d_out"],
            roughness=pipe_props["roughness"],
            conductivity_inner=pipe_props["conductivity_inner"],
            conductivity_outer=pipe_props["conductivity_outer"],
            rho_cp=pipe_props["rho_cp"]
        )

    ghe.set_borehole(
        height=constraint_props["max_height"],
        buried_depth=borehole_props["buried_depth"],
        diameter=borehole_props["diameter"]
    )

    ghe.set_ground_loads_from_hourly_list(ground_load_props)
    max_bh = design_props["max_boreholes"] if "max_boreholes" in design_props.keys() else None
    continue_if_design_unmet = design_props[
        "continue_if_design_unmet"] if "continue_if_design_unmet" in design_props.keys() else False
    ghe.set_simulation_parameters(
        num_months=sim_props["num_months"],
        max_eft=design_props["max_eft"],
        min_eft=design_props["min_eft"],
        max_height=constraint_props["max_height"],
        min_height=constraint_props["min_height"],
        max_boreholes=max_bh,
        continue_if_design_unmet=continue_if_design_unmet
    )

    if ghe.set_design_geometry_type(constraint_props["method"], throw=False) != 0:
        return 1

    if ghe.geom_type == DesignGeomType.RECTANGLE:
        ghe.set_geometry_constraints_rectangle(
            length=constraint_props["length"],
            width=constraint_props["width"],
            b_min=constraint_props["b_min"],
            b_max=constraint_props["b_max"],
        )
    elif ghe.geom_type == DesignGeomType.NEARSQUARE:
        ghe.set_geometry_constraints_near_square(
            b=constraint_props["b"],
            length=constraint_props["length"]
        )
    elif ghe.geom_type == DesignGeomType.BIRECTANGLE:
        ghe.set_geometry_constraints_bi_rectangle(
            length=constraint_props["length"],
            width=constraint_props["width"],
            b_min=constraint_props["b_min"],
            b_max_x=constraint_props["b_max_x"],
            b_max_y=constraint_props["b_max_y"]
        )
    elif ghe.geom_type == DesignGeomType.BIZONEDRECTANGLE:
        ghe.set_geometry_constraints_bi_zoned_rectangle(
            length=constraint_props["length"],
            width=constraint_props["width"],
            b_min=constraint_props["b_min"],
            b_max_x=constraint_props["b_max_x"],
            b_max_y=constraint_props["b_max_y"]
        )
    elif ghe.geom_type == DesignGeomType.BIRECTANGLECONSTRAINED:
        ghe.set_geometry_constraints_bi_rectangle_constrained(
            b_min=constraint_props["b_min"],
            b_max_x=constraint_props["b_max_x"],
            b_max_y=constraint_props["b_max_y"],
            property_boundary=constraint_props["property_boundary"],
            no_go_boundaries=constraint_props["no_go_boundaries"]
        )
    elif ghe.geom_type == DesignGeomType.ROWWISE:
        # We are using perimeter calculations if present
        if "perimeter_spacing_ratio" in constraint_props.keys():
            perimeter_spacing_ratio = constraint_props["perimeter_spacing_ratio"]
        else:
            perimeter_spacing_ratio = None

        ghe.set_geometry_constraints_rowwise(
            perimeter_spacing_ratio=perimeter_spacing_ratio,
            max_spacing=constraint_props["max_spacing"],
            min_spacing=constraint_props["min_spacing"],
            spacing_step=constraint_props["spacing_step"],
            max_rotation=constraint_props["max_rotation"],
            min_rotation=constraint_props["min_rotation"],
            rotate_step=constraint_props["rotate_step"],
            property_boundary=constraint_props["property_boundary"],
            no_go_boundaries=constraint_props["no_go_boundaries"]
        )
    else:
        print("Geometry constraint method not supported.", file=stderr)
        return 1

    ghe.set_design(
        flow_rate=design_props["flow_rate"],
        flow_type_str=design_props["flow_type"],
        throw=False
    )

    ghe.find_design(throw=False)
    ghe.prepare_results("GHEDesigner Run from CLI", "Notes", "Author", "Iteration Name")
    ghe.write_output_files(output_directory)

    return 0


@click.command(name="GHEDesignerCommandLine")
@click.argument("input-path", type=click.Path(exists=True), required=True)
@click.argument("output-directory", type=click.Path(exists=False), required=False)
@click.version_option(VERSION)
@click.option(
    "--validate",
    default=False,
    is_flag=True,
    show_default=False,
    help="Validate input and exit."
)
@click.option(
    "-c",
    "--convert",
    help="Convert output to specified format. Options supported: 'IDF'."
)
def run_manager_from_cli(input_path, output_directory, validate, convert):
    input_path = Path(input_path).resolve()

    if validate:
        try:
            validate_input_file(input_path)
            print("Valid input file.")
            return 0
        except ValidationError:
            print("Schema validation error. See previous error message for details.", file=stderr)
            return 1

    if convert:
        if convert == "IDF":
            try:
                write_idf(input_path)
                print("Output converted to IDF objects.")
                return 0
            except:
                print("Conversion to IDF error.", file=stderr)
                return 1

        else:
            print(f"Unsupported conversion format type: {format}", file=stderr)
            return 1

    if output_directory is None:
        print('Output directory path must be passed as an argument, aborting', file=stderr)
        return 1

    output_path = Path(output_directory).resolve()

    return run_manager_from_cli_worker(input_path, output_path)


if __name__ == "__main__":
    exit(run_manager_from_cli())



class SimulationParameters:
    def __init__(
            self,
            start_month,
            end_month,
            max_entering_fluid_temp_allow,
            min_entering_fluid_temp_allow,
            max_height,
            min_height,
            max_boreholes=None,
            continue_if_design_unmet=False
    ):
        # Simulation parameters not found in other objects
        # ------------------------------------------------
        # Simulation start month and end month
        self.start_month = start_month
        self.end_month = end_month
        # Maximum and minimum allowable fluid temperatures
        self.max_EFT_allowable = max_entering_fluid_temp_allow  # degrees Celsius
        self.min_EFT_allowable = min_entering_fluid_temp_allow  # degrees Celsius
        # Maximum and minimum allowable heights
        self.max_height = max_height  # in meters
        self.min_height = min_height  # in meters
        self.max_boreholes = max_boreholes
        self.continue_if_design_unmet = continue_if_design_unmet

    def as_dict(self) -> dict:
        output = dict()
        output['type'] = str(self.__class__)
        output['start_month'] = self.start_month
        output['end_month'] = self.end_month
        output['max_eft_allowable'] = {'value': self.max_EFT_allowable, 'units': 'C'}
        output['min_eft_allowable'] = {'value': self.min_EFT_allowable, 'units': 'C'}
        output['maximum_height'] = {'value': self.max_height, 'units': 'm'}
        output['minimum_height'] = {'value': self.min_height, 'units': 'm'}

        if self.max_boreholes is not None:
            output['maximum_boreholes'] = {'value': self.max_boreholes, 'units': '-'}
        return output

    def to_input(self) -> dict:
        return {'num_months': self.end_month}



from typing import List, Tuple, Union


def transpose_coordinates(coordinates) -> List[Tuple[float, float]]:
    coordinates_transposed = []
    for x, y in coordinates:
        coordinates_transposed.append((y, x))
    return coordinates_transposed


def rectangle(num_bh_x: int, num_bh_y: int, spacing_x: Union[int, float], spacing_y: Union[int, float],
              origin=(0, 0), ) -> List[Tuple[float, float]]:
    """
    Creates a rectangular borehole field.

    X   X   X   X
    X   X   X   X
    X   X   X   X
    X   X   X   X

    Args:
        num_bh_x: number of borehole rows in x-direction
        num_bh_y: number of borehole rows in y-direction
        spacing_x: spacing between borehole rows in x-direction
        spacing_y: spacing between borehole rows in y-direction
        origin: coordinates for origin at lower-left corner

    Returns:
        list of tuples (x, y) containing borehole coordinates
    """

    r = []
    x_0 = origin[0]
    y_0 = origin[1]
    for i in range(num_bh_x):
        for j in range(num_bh_y):
            r.append((x_0 + i * spacing_x, y_0 + j * spacing_y))

    return r


def open_rectangle(num_bh_x: int, num_bh_y: int, spacing_x: Union[int, float],
                   spacing_y: Union[int, float]) -> List[Tuple[float, float]]:
    """
    Creates a rectangular borehole field without center boreholes.

    X   X   X   X
    X           X
    X           X
    X   X   X   X

    Args:
        num_bh_x: number of borehole rows in x-direction
        num_bh_y: number of borehole rows in y-direction
        spacing_x: spacing between borehole rows in x-direction
        spacing_y: spacing between borehole rows in y-direction

    Returns:
        list of tuples (x, y) containing borehole coordinates
    """

    open_r = []
    if num_bh_x > 2 and num_bh_y > 2:
        for i in range(num_bh_x):
            open_r.append((i * spacing_x, 0.0))
        for j in range(1, num_bh_y - 1):
            open_r.append((0, j * spacing_y))
            open_r.append(((num_bh_x - 1) * spacing_x, j * spacing_y))
        for i in range(num_bh_x):
            open_r.append((i * spacing_x, (num_bh_y - 1) * spacing_y))
        # nbh = num_bh_y * 2 + (num_bh_x - 2) * 2
    else:
        open_r = rectangle(num_bh_x, num_bh_y, spacing_x, spacing_y)
        # nbh = num_bh_x * num_bh_y

    return open_r


def c_shape(n_x_1: int, n_y: int, b_x: Union[int, float],
            b_y: Union[int, float], n_x_2: int) -> List[Tuple[float, float]]:
    c = []
    for i in range(n_x_1):
        c.append((i * b_x, 0.0))
    x_loc = (n_x_1 - 1) * b_x
    for j in range(1, n_y):
        c.append((0.0, j * b_y))
    for j in range(1, n_y):
        c.append((x_loc, j * b_y))
    y_loc = (n_y - 1) * b_y
    for i in range(1, n_x_2 + 1):
        c.append((i * b_x, y_loc))

    return c


def lop_u(n_x: int, n_y_1: int, b_x: Union[int, float],
          b_y: Union[int, float], n_y_2: int) -> List[Tuple[float, float]]:
    _lop_u = []
    for i in range(n_x):
        _lop_u.append((i * b_x, 0.0))
    for j in range(1, n_y_1):
        _lop_u.append((0.0, j * b_y))
    x_loc = (n_x - 1) * b_x
    for j in range(1, n_y_2):
        _lop_u.append((x_loc, j * b_y))

    return _lop_u


def l_shape(n_x: int, n_y: int, b_x: Union[int, float], b_y: Union[int, float]) -> List[Tuple[float, float]]:
    l_shape_object = []
    for i in range(n_x):
        l_shape_object.append((i * b_x, 0.0))
    for j in range(1, n_y):
        l_shape_object.append((0.0, j * b_y))

    return l_shape_object


def zoned_rectangle(n_x: int, n_y: int, b_x: Union[int, float], b_y: Union[int, float],
                    n_ix: int, n_iy: int) -> List[Tuple[float, float]]:
    """
    Create a zoned rectangle

    :param n_x:
    :param n_y:
    :param b_x:
    :param b_y:
    :param n_ix:
    :param n_iy:
    :return:
    """

    if n_ix > (n_x - 2):
        raise ValueError("To many interior x boreholes.")
    if n_iy > (n_y - 2):
        raise ValueError("Too many interior y boreholes.")

    # Create a list of (x, y) coordinates
    zoned = []

    # Boreholes on the perimeter
    zoned.extend(open_rectangle(n_x, n_y, b_x, b_y))

    # Create the interior coordinates
    bix = (n_x - 1) * b_x / (n_ix + 1)
    biy = (n_y - 1) * b_y / (n_iy + 1)

    zoned.extend(rectangle(n_ix, n_iy, bix, biy, origin=(bix, biy)))

    return zoned



import warnings
from calendar import monthrange
from json import dumps
from math import floor

import numpy as np
from scipy.interpolate import interp1d

from ghedesigner.borehole_heat_exchangers import SingleUTube
from ghedesigner.constants import TWO_PI
from ghedesigner.radial_numerical_borehole import RadialNumericalBH
from ghedesigner.simulation import SimulationParameters


class HybridLoad:
    def __init__(
            self,
            raw_loads: list,
            bhe: SingleUTube,
            radial_numerical: RadialNumericalBH,
            sim_params: SimulationParameters,
            years=None,
    ):
        # Split the hourly loads into heating and cooling (kW)
        if years is None:
            years = [2019]

        self.hourly_rejection_loads, self.hourly_extraction_loads = self.split_heat_and_cool(raw_loads)

        # Simulation start and end month
        self.start_month = sim_params.start_month
        self.end_month = sim_params.end_month
        if len(years) <= 1:
            self.peak_retain_start = 12  # use peak loads for first 12 months
            self.peak_retain_end = 12  # use peak loads for last 12 months
        else:
            self.peak_retain_start = len(years) * 6
            self.peak_retain_end = len(years) * 6

        # Store the borehole heat exchanger
        self.bhe = bhe
        # Store the radial numerical g-function value
        # Note: this is intended to be a scipy.interp1d object
        self.radial_numerical = radial_numerical
        self.years = years

        # Get the number of days in each month for a given year (make 0 NULL)
        self.days_in_month = [0]
        for year in years:
            self.days_in_month.extend([monthrange(year, i)[1] for i in range(1, 13)])

        # TODO: verify whether errors are possible here and raise exception if needed
        # assert (len(hourly_rejection_loads) == sum(self.days_in_month) * 24.0
        #         and len(hourly_extraction_loads) == sum(self.days_in_month) * 24.0), (
        #     "The total number of hours in the year are not equal. Is this a leap year?")

        # This block of data holds the compact monthly representation of the
        # loads. The intention is that these loads will usually repeat. It's
        # possible that for validation or design purposes, users may wish to
        # specify loads that differ from year to year. For these arrays,
        # January is the second item (1) and December the last (12)
        # We'll reserve the first item (0) for an annual total or peak

        num_unique_months = len(years) * 12 + 1

        # monthly cooling loads (or heat rejection) in kWh
        self.monthly_cl = [0] * num_unique_months
        # monthly heating loads (or heat extraction) in kWh
        self.monthly_hl = [0] * num_unique_months
        # monthly peak cooling load (or heat rejection) in kW
        self.monthly_peak_cl = [0] * num_unique_months
        # monthly peak heating load (or heat extraction) in kW
        self.monthly_peak_hl = [0] * num_unique_months
        # monthly average cooling load (or heat rejection) in kW
        self.monthly_avg_cl = [0] * num_unique_months
        # monthly average heating load (or heat extraction) in kW
        self.monthly_avg_hl = [0] * num_unique_months
        # day of the month on which peak clg load occurs (e.g. 1-31)
        self.monthly_peak_cl_day = [0] * num_unique_months
        # day of the month on which peak htg load occurs (e.g. 1-31)
        self.monthly_peak_hl_day = [0] * num_unique_months
        # Process the loads by month
        self.split_loads_by_month()

        # 48 hour loads are going to be necessary for the hourly simulation for
        # finding the peak load duration
        # These will be a 2D list, a list of 48 hour loads in each index
        # Make 0 position NULL
        # list of two day (48 hour) cooling loads (or heat rejection) in kWh
        self.two_day_hourly_peak_cl_loads = [[0]]
        # list of two day (48 hour) heating loads (or heat extraction) in kWh
        self.two_day_hourly_peak_hl_loads = [[0]]
        self.process_two_day_loads()

        # Now we need to perform 48-hour simulations to determine the
        # monthly peak load hours
        # Stores two day (48 hour) fluid temperatures for cooling with nominal
        # load
        self.two_day_fluid_temps_cl_nm = [[0]]
        # Stores two day (48 hour) fluid temperatures for cooling with peak load
        self.two_day_fluid_temps_cl_pk = [[0]]
        # Stores two day (48 hour) fluid temperatures for heating with nominal
        # load
        self.two_day_fluid_temps_hl_nm = [[0]]
        # Stores two day (48 hour) fluid temperatures for heating with peak load
        self.two_day_fluid_temps_hl_pk = [[0]]

        # duration of monthly peak clg load in hours
        self.monthly_peak_cl_duration = [0] * num_unique_months
        # duration of monthly peak htg load in hours
        self.monthly_peak_hl_duration = [0] * num_unique_months
        self.find_peak_durations()

        # This block of data holds the sequence of loads. This is an
        # intermediate form, where the load values hold the actual loads,
        # not the de-convoluted loads
        self.load = np.array(0)  # holds the load during the period
        self.hour = np.array(0)  # holds the last hour of a period
        self.step_func_load = np.array(0)  # holds the load in terms of step functions
        self.process_month_loads()

    def as_dict(self) -> dict:
        output = dict()
        output['type'] = str(self.__class__)
        output['results'] = self.create_dataframe_of_peak_analysis()
        return output

    @staticmethod
    def split_heat_and_cool(raw_loads):
        """
        Split the provided loads into heating and cooling.
        Heating is positive, cooling is negative.

        :param raw_loads: raw loads entered by the user, in Watts
        :return: Loads split into heating and cooling
        """
        hourly_extraction_loads = [x / 1000.0 if x >= 0.0 else 0.0 for x in raw_loads]
        hourly_rejection_loads = [abs(x) / 1000.0 if x < 0.0 else 0.0 for x in raw_loads]

        return hourly_rejection_loads, hourly_extraction_loads

    def split_loads_by_month(self) -> None:
        # Split the loads into peak, total and average loads for each month

        hours_in_day = 24
        # Store the index of the last months hours
        hours_in_previous_months = 0
        for i in range(1, len(self.days_in_month)):
            hours_in_month = hours_in_day * self.days_in_month[i]
            # Slice the hours in this current month
            month_rejection_loads = \
                self.hourly_rejection_loads[hours_in_previous_months: hours_in_previous_months + hours_in_month]
            month_extraction_loads = \
                self.hourly_extraction_loads[hours_in_previous_months: hours_in_previous_months + hours_in_month]

            # TODO: verify whether errors are possible here and raise exception if needed
            # assert (len(month_extraction_loads) == hours_in_month and len(month_rejection_loads) == hours_in_month)

            # Sum
            # monthly cooling loads (or heat rejection) in kWh
            self.monthly_cl[i] = sum(month_rejection_loads)
            # monthly heating loads (or heat extraction) in kWh
            self.monthly_hl[i] = sum(month_extraction_loads)

            # Peak
            # monthly peak cooling load (or heat rejection) in kW
            self.monthly_peak_cl[i] = max(month_rejection_loads)
            # monthly peak heating load (or heat extraction) in kW
            self.monthly_peak_hl[i] = max(month_extraction_loads)

            # Average
            # monthly average cooling load (or heat rejection) in kW
            self.monthly_avg_cl[i] = self.monthly_cl[i] / len(month_rejection_loads)
            # monthly average heating load (or heat extraction) in kW
            self.monthly_avg_hl[i] = self.monthly_hl[i] / len(month_extraction_loads)

            # Day of month the peak heating load occurs
            # day of the month on which peak clg load occurs (e.g. 1-31)
            self.monthly_peak_cl_day[i] = floor(month_rejection_loads.index(self.monthly_peak_cl[i]) / hours_in_day)
            # day of the month on which peak clg load occurs (e.g. 1-31)
            self.monthly_peak_hl_day[i] = floor(month_extraction_loads.index(self.monthly_peak_hl[i]) / hours_in_day)
            # print("Monthly Peak HL Hour",month_extraction_loads.index(
            # self.monthly_peak_hl[i]) / hours_in_day)
            # print("Monthly Peak HL Day: ",self.monthly_peak_hl_day[i])
            # print("")

            hours_in_previous_months += hours_in_month

    def process_two_day_loads(self) -> None:
        # The two day (48 hour) two day loads are selected by locating the day
        # the peak load of the month occurs on, and pulling a 48-hour load
        # profile -- the day before and the day of

        hours_in_day = 24
        hours_in_year = len(self.hourly_rejection_loads)

        # Add the last day of the year to the beginning of the loads to account
        # for the possibility that a peak load occurs on the first day of the
        # year

        hourly_rejection_loads = (self.hourly_rejection_loads[
                                  hours_in_year - hours_in_day: hours_in_year] + self.hourly_rejection_loads)
        hourly_extraction_loads = (self.hourly_extraction_loads[
                                   hours_in_year - hours_in_day: hours_in_year] + self.hourly_extraction_loads)

        # Keep track of how many hours are in
        # start at 24 since we added the last day of the year to the beginning
        hours_in_previous_months = hours_in_day
        # loop over all 12 months
        for i in range(1, len(self.days_in_month)):
            hours_in_month = hours_in_day * self.days_in_month[i]

            # day of the month on which peak clg load occurs (e.g. 1-31)
            monthly_peak_cl_day = self.monthly_peak_cl_day[i]
            # day of the month on which peak clg load occurs (e.g. 1-31)
            monthly_peak_hl_day = self.monthly_peak_hl_day[i]
            # Get the starting hour of the day before the peak cooling load day
            monthly_peak_cl_hour_start = (hours_in_previous_months + (monthly_peak_cl_day - 1) * hours_in_day)
            # Get the starting hour of the day before the peak heating load day
            monthly_peak_hl_hour_start = (hours_in_previous_months + (monthly_peak_hl_day - 1) * hours_in_day)

            # monthly cooling loads (or heat rejection) in kWh
            two_day_hourly_peak_cl_load = \
                hourly_rejection_loads[monthly_peak_cl_hour_start: monthly_peak_cl_hour_start + 2 * hours_in_day]
            # monthly heating loads (or heat extraction) in kWh
            two_day_hourly_peak_hl_load = \
                hourly_extraction_loads[monthly_peak_hl_hour_start: monthly_peak_hl_hour_start + 2 * hours_in_day]

            # TODO: verify whether errors are possible here and raise exception if needed
            # assert (len(two_day_hourly_peak_hl_load) == 2 * hours_in_day
            #         and len(two_day_hourly_peak_cl_load) == 2 * hours_in_day)

            # Double check ourselves
            # TODO: verify whether errors are possible here and raise exception if needed
            # monthly_peak_cl_day_start = int((monthly_peak_cl_hour_start - hours_in_day) / hours_in_day)
            # monthly_peak_cl_hour_month = int(monthly_peak_cl_day_start - sum(self.days_in_month[0:i]))
            # assert monthly_peak_cl_hour_month == monthly_peak_cl_day - 1

            # TODO: verify whether errors are possible here and raise exception if needed
            # monthly_peak_hl_day_start = (monthly_peak_hl_hour_start - hours_in_day) / hours_in_day
            # monthly_peak_hl_hour_month = int(monthly_peak_hl_day_start - sum(self.days_in_month[0:i]))
            # assert monthly_peak_hl_hour_month == monthly_peak_hl_day - 1

            # monthly cooling loads (or heat rejection) in kWh
            self.two_day_hourly_peak_cl_loads.append(two_day_hourly_peak_cl_load)
            # monthly heating loads (or heat extraction) in kWh
            self.two_day_hourly_peak_hl_loads.append(two_day_hourly_peak_hl_load)

            hours_in_previous_months += hours_in_month

    @staticmethod
    def simulate_hourly(hour_time, q, g_sts, resist_bh, two_pi_k, ts):
        # An hourly simulation for the fluid temperature
        # Chapter 2 of Advances in Ground Source Heat Pumps

        q_dt = np.hstack((q[1:] - q[:-1]))

        delta_t_fluid = [0]
        for n in range(1, len(hour_time)):
            # Take the last i elements of the reversed time array
            _time = hour_time[n] - hour_time[0:n]
            # _time = time_values_reversed[n - i:n]
            g_values = g_sts(np.log((_time * 3600.0) / ts))
            # Tb = Tg + (q_dt * g)  (Equation 2.12)
            delta_tb_i = (q_dt[0:n] / two_pi_k).dot(g_values)
            # Delta mean heat pump entering fluid temperature
            tf_mean = delta_tb_i + q[n] * resist_bh
            delta_t_fluid.append(tf_mean)

        return delta_t_fluid

    def perform_current_month_simulation(
            self,
            two_day_hourly_peak_load,
            peak_load,
            avg_load,
            two_day_fluid_temps_pk,
            two_day_fluid_temps_nm,
    ):
        ts = self.radial_numerical.t_s
        two_pi_k = TWO_PI * self.bhe.soil.k
        resist_bh_effective = self.bhe.calc_effective_borehole_resistance()
        g_sts = self.radial_numerical.g_sts
        hours_in_day = 24
        hour_time = np.array(list(range(0, 2 * hours_in_day + 1)))
        # Two day peak cooling load scaled down by average (q_max - q_avg)
        q_peak = np.array([0.0] + [peak_load - avg_load] * (2 * hours_in_day))
        # Two day nominal cooling load (q_i - q_avg) / q_max * q_i
        q_nominal = np.array(
            [0.0]
            + [
                (two_day_hourly_peak_load[i] - avg_load)
                / peak_load
                * two_day_hourly_peak_load[i]
                for i in range(1, len(q_peak))
            ]
        )
        # Get peak fluid temperatures using peak load
        delta_t_fluid_peak = self.simulate_hourly(hour_time, q_peak, g_sts, resist_bh_effective, two_pi_k, ts)
        two_day_fluid_temps_pk.append(delta_t_fluid_peak)
        # Get nominal fluid temperatures using nominal load
        delta_t_fluid_nom = self.simulate_hourly(hour_time, q_nominal, g_sts, resist_bh_effective, two_pi_k, ts)
        two_day_fluid_temps_nm.append(delta_t_fluid_nom)

        delta_t_fluid_nom_max = max(delta_t_fluid_nom)

        if delta_t_fluid_nom_max > 0.0:
            f = interp1d(delta_t_fluid_peak, hour_time)
            peak_duration = f(delta_t_fluid_nom_max).tolist()
        else:
            peak_duration = 1.0e-6

        return peak_duration, q_peak, q_nominal

    def find_peak_durations(self) -> None:
        # Find the peak durations using hourly simulations for 2 days

        for i in range(1, len(self.days_in_month)):
            # Scale all the loads by the peak load
            # Perform an hourly simulation with the scaled loads
            # Perform an hourly simulation with a load of 1, or the peak loads
            # divided by the peak

            # two day cooling loads (or heat rejection) in kWh
            current_two_day_cl_load = [0.0] + self.two_day_hourly_peak_cl_loads[i]

            # This tolerance applies to the difference between the current
            # months peak load and the maximum of the two-day load. If the
            # absolute value of the difference between the current months
            # peak load and the current two-day peak load is within this
            # tolerance, then the maximum of the two-day load is equal to the
            # maximum of the current month. If the absolute difference is
            # greater than the tolerance, then the two-day peak load contains
            # a load greater than the current months peak load. The tolerance
            # could ONLY be exceeded when the first 24 hours is located in the
            # previous month.
            tol = 0.1

            # Ensure the peak load for the two-day load profile is the same or
            # greater than the monthly peak load. This check is done in case
            # the previous month contains a higher load than the current month.
            load_diff = self.monthly_peak_cl[i] - max(current_two_day_cl_load)
            # monthly peak cooling load (or heat rejection) in kW
            if abs(load_diff) < tol:
                current_month_peak_cl = self.monthly_peak_cl[i]
            else:
                current_month_peak_cl = max(current_two_day_cl_load)

            # monthly average cooling load (or heat rejection) in kW
            current_month_avg_cl = self.monthly_avg_cl[i]

            if current_month_peak_cl != 0.0:
                peak_duration, q_peak, q_nominal = self.perform_current_month_simulation(
                    current_two_day_cl_load,
                    current_month_peak_cl,
                    current_month_avg_cl,
                    self.two_day_fluid_temps_cl_pk,
                    self.two_day_fluid_temps_cl_nm)
            else:
                peak_duration = 1.0e-6

            # Set the monthly cooling load duration
            self.monthly_peak_cl_duration[i] = peak_duration

            # two day heating loads (or heat extraction) in kWh
            current_two_day_hl_load = [0.0] + self.two_day_hourly_peak_hl_loads[i]

            # Ensure the peak load for the two-day load profile is the same or
            # greater than the monthly peak load. This check is done in case
            # the previous month contains a higher load than the current month.
            load_diff = self.monthly_peak_hl[i] - max(current_two_day_hl_load)
            # monthly peak cooling load (or heat rejection) in kW
            if abs(load_diff) < tol:
                current_month_peak_hl = self.monthly_peak_hl[i]
            else:
                current_month_peak_hl = max(current_two_day_hl_load)

            # monthly average heating load (or heat extraction) in kW
            current_month_avg_hl = self.monthly_avg_hl[i]

            if current_month_peak_hl != 0.0:
                peak_duration, q_peak, q_nominal = self.perform_current_month_simulation(
                    current_two_day_hl_load,
                    current_month_peak_hl,
                    current_month_avg_hl,
                    self.two_day_fluid_temps_hl_pk,
                    self.two_day_fluid_temps_hl_nm,
                )
            else:
                peak_duration = 1.0e-6

            # Set the monthly cooling load duration
            self.monthly_peak_hl_duration[i] = peak_duration

    def create_dataframe_of_peak_analysis(self) -> str:
        # The fields are: sum, peak, avg, peak day, peak duration
        hybrid_time_step_fields = {
            "Total": {},
            "Peak": {},
            "Average": {},
            "Peak Day": {},
            "Peak Duration": {},
        }

        d: dict = {}
        # For all the months, create dictionary of fields
        for i in range(1, 13):
            m_n = number_to_month(i)
            d[m_n] = hybrid_time_step_fields

            # set total
            d[m_n]["Total"]["rejection"] = self.monthly_cl[i]
            d[m_n]["Total"]["extraction"] = self.monthly_hl[i]
            # set peak
            d[m_n]["Peak"]["rejection"] = self.monthly_peak_cl[i]
            d[m_n]["Peak"]["extraction"] = self.monthly_peak_hl[i]
            # set average
            d[m_n]["Average"]["rejection"] = self.monthly_avg_cl[i]
            d[m_n]["Average"]["extraction"] = self.monthly_avg_hl[i]
            # set peak day
            d[m_n]["Peak Day"]["rejection"] = self.monthly_peak_cl_day[i]
            d[m_n]["Peak Day"]["extraction"] = self.monthly_peak_hl_day[i]
            # set peak duration
            d[m_n]["Peak Duration"]["rejection"] = self.monthly_peak_cl_duration[i]
            d[m_n]["Peak Duration"]["extraction"] = self.monthly_peak_hl_duration[i]

        return dumps(d, indent=2)

    def process_month_loads(self):
        # Converts monthly load format to sequence of loads needed for
        # simulation
        # This routine is taking loads applied to the ground NOT to a heat pump.

        warn_msg_neg_timestep = "A negative time step has been generated in the hybrid loading scheme. \n" \
                                "This will reduce the accuracy of the simulation."

        # First, begin array with zero load before simulation starts.
        self.load = np.append(self.load, 0)
        last_zero_hour = first_month_hour(self.start_month, self.years) - 1
        self.hour = np.append(self.hour, last_zero_hour)
        if len(self.years) <= 1:
            # Second, replicate months. [if we want to add an option where all
            # monthly loads are explicitly given, this code will be in an if block]
            for i in range(self.start_month, self.end_month + 1):
                if i > 12:
                    mi = i % 12
                    if mi == 0:
                        mi = 12
                    self.monthly_cl.append(self.monthly_cl[mi])
                    self.monthly_hl.append(self.monthly_hl[mi])
                    self.monthly_peak_cl.append(self.monthly_peak_cl[mi])
                    self.monthly_peak_hl.append(self.monthly_peak_hl[mi])
                    self.monthly_peak_cl_duration.append(self.monthly_peak_cl_duration[mi])
                    self.monthly_peak_hl_duration.append(self.monthly_peak_hl_duration[mi])
                    self.monthly_peak_cl_day.append(self.monthly_peak_cl_day[mi])
                    self.monthly_peak_hl_day.append(self.monthly_peak_hl_day[mi])
        # Set the ipf (include peak flag)
        if len(self.years) <= 1:
            ipf = [False] * (self.end_month + 1)
            for i in range(self.start_month, self.end_month + 1):
                # set flag that determines if peak load will be included
                if i < self.start_month + self.peak_retain_start:
                    ipf[i] = True
                if i > self.end_month - self.peak_retain_end:
                    ipf[i] = True
        else:
            ipf = [True] * (self.end_month + 1)
        pass
        peak_last_avg_hour = 0.0
        for i in range(self.start_month, (self.end_month + 1)):
            # There may be a more sophisticated way to do this, but I will loop
            # through the lists month duration is the number of hours over which to
            # calculate the average value for the month
            if ipf[i]:
                if len(self.years) <= 1:
                    current_year = self.years[0]
                else:
                    current_year = self.years[(i - 1) // 12]
                month_duration = \
                    monthdays(i, current_year) * 24 - self.monthly_peak_cl_duration[i] - \
                    self.monthly_peak_hl_duration[i]
                # gives htg load pk energy in kWh
                month_peak_hl = self.monthly_peak_hl[i] * self.monthly_peak_hl_duration[i]
                # gives htg load pk energy in kWh
                month_peak_cl = self.monthly_peak_cl[i] * self.monthly_peak_cl_duration[i]
                month_load = self.monthly_cl[i] - self.monthly_hl[i] - month_peak_cl + month_peak_hl
                month_rate = month_load / month_duration
                peak_day_diff = self.monthly_peak_cl_day[i] - self.monthly_peak_hl_day[i]
                # Place the peaks roughly midway through the day they occur on.
                # (In JDS's opinion, this should be amply accurate for the
                # hybrid time step.)
                # Catch the first and last peak hours to make sure they aren't 0
                # Could only be 0 when the first month has no load.
                first_hour_heating_peak = \
                    first_month_hour(i, self.years) + \
                    (self.monthly_peak_hl_day[i]) * 24 + 12 - (self.monthly_peak_hl_duration[i] / 2)
                if first_hour_heating_peak < 0.0:
                    first_hour_heating_peak = 1.0e-6
                last_hour_heating_peak = first_hour_heating_peak + self.monthly_peak_hl_duration[i]
                if last_hour_heating_peak < 0.0:
                    last_hour_heating_peak = 1.0e-6
                first_hour_cooling_peak = \
                    first_month_hour(i, self.years) + (self.monthly_peak_cl_day[i]) * 24 + 12 - \
                    self.monthly_peak_cl_duration[i] / 2
                if first_hour_cooling_peak < 0.0:
                    first_hour_cooling_peak = 1.0e-06
                last_hour_cooling_peak = first_hour_cooling_peak + self.monthly_peak_cl_duration[i]
                if last_hour_cooling_peak < 0.0:
                    last_hour_cooling_peak = 1.0e-06
            else:  # peak load not used this month

                month_duration = monthdays(i, current_year) * 24

                month_load = self.monthly_cl[i] - self.monthly_hl[i]
                month_rate = month_load / month_duration
                peak_day_diff = 0

            last_avg_hour = 0.0
            if peak_day_diff < 0:
                # monthly peak heating day occurs after peak cooling day
                # monthly average conditions before cooling peak
                if self.monthly_peak_cl[i] > 0 and ipf[i]:
                    # last_avg_hour = first_hour_cooling_peak - 1 JDS corrected 20200604
                    last_avg_hour = first_hour_cooling_peak
                    self.load = np.append(self.load, month_rate)
                    self.hour = np.append(self.hour, last_avg_hour)
                    # cooling peak
                    # self.load = np.append(self.load, -self.monthly_peak_cl[i]) JDS corrected 20200604
                    self.load = np.append(self.load, self.monthly_peak_cl[i])
                    self.hour = np.append(self.hour, last_hour_cooling_peak)

                    if last_avg_hour - peak_last_avg_hour < 0.0:
                        warnings.warn(warn_msg_neg_timestep)
                    peak_last_avg_hour = last_avg_hour
                # monthly average conditions between cooling peak and heating peak
                if self.monthly_peak_hl[i] > 0 and ipf[i]:
                    # last_avg_hour = first_hour_heating_peak - 1 JDS corrected 20200604
                    last_avg_hour = first_hour_heating_peak
                    self.load = np.append(self.load, month_rate)
                    self.hour = np.append(self.hour, last_avg_hour)
                    # heating peak
                    # self.load = np.append(self.load, self.monthly_peak_hl[i]) JDS corrected 20200604
                    self.load = np.append(self.load, -self.monthly_peak_hl[i])
                    self.hour = np.append(self.hour, last_hour_heating_peak)

                    if last_avg_hour - peak_last_avg_hour < 0.0:
                        warnings.warn(warn_msg_neg_timestep)
                    peak_last_avg_hour = last_avg_hour
                # rest of month
                last_avg_hour = last_month_hour(i, self.years)
                self.load = np.append(self.load, month_rate)
                self.hour = np.append(self.hour, last_avg_hour)

                if last_avg_hour - peak_last_avg_hour < 0.0:
                    warnings.warn(warn_msg_neg_timestep)
                peak_last_avg_hour = last_avg_hour

            elif peak_day_diff > 0:
                # monthly peak heating day occurs before peak cooling day
                # monthly average conditions before cooling peak
                if self.monthly_peak_hl[i] > 0 and ipf[i]:
                    last_avg_hour = first_hour_heating_peak
                    self.load = np.append(self.load, month_rate)
                    self.hour = np.append(self.hour, last_avg_hour)
                    # heating peak
                    self.load = np.append(self.load, -self.monthly_peak_hl[i])
                    self.hour = np.append(self.hour, last_hour_heating_peak)

                    if last_avg_hour - peak_last_avg_hour < 0.0:
                        warnings.warn(warn_msg_neg_timestep)
                    peak_last_avg_hour = last_avg_hour
                # monthly average conditions between heating peak and cooling peak
                if self.monthly_peak_cl[i] > 0 and ipf[i]:
                    last_avg_hour = first_hour_cooling_peak
                    self.load = np.append(self.load, month_rate)
                    self.hour = np.append(self.hour, last_avg_hour)
                    # cooling peak
                    self.load = np.append(self.load, self.monthly_peak_cl[i])
                    self.hour = np.append(self.hour, last_hour_cooling_peak)

                    if last_avg_hour - peak_last_avg_hour < 0.0:
                        warnings.warn(warn_msg_neg_timestep)
                    peak_last_avg_hour = last_avg_hour
                # rest of month
                last_avg_hour = last_month_hour(i, self.years)
                self.load = np.append(self.load, month_rate)
                self.hour = np.append(self.hour, last_avg_hour)

                if last_avg_hour - peak_last_avg_hour < 0.0:
                    warnings.warn(warn_msg_neg_timestep)
                peak_last_avg_hour = last_avg_hour
            else:
                # monthly peak heating day and cooling day are the same
                # Cooling Load placed before noon, and the heating load is placed after noon
                # Currently the exact times of the heating and cooling peaks are not stored. If further work is done
                # this default can be made to be more accurate.
                if ipf[i]:
                    # monthly average conditions before cooling peak
                    if self.monthly_peak_cl[i] > 0 and ipf[i]:
                        # last_avg_hour = first_hour_cooling_peak - 1 JDS corrected 20200604
                        last_avg_hour = first_hour_cooling_peak - self.monthly_peak_cl_duration[i] / 2
                        self.load = np.append(self.load, month_rate)
                        self.hour = np.append(self.hour, last_avg_hour)
                        # cooling peak
                        # self.load = np.append(self.load, -self.monthly_peak_cl[i]) JDS corrected 20200604
                        self.load = np.append(self.load, self.monthly_peak_cl[i])
                        self.hour = np.append(
                            self.hour,
                            last_hour_cooling_peak
                            - self.monthly_peak_cl_duration[i] / 2,
                        )

                        if last_avg_hour - peak_last_avg_hour < 0.0:
                            warnings.warn(warn_msg_neg_timestep)
                        peak_last_avg_hour = last_avg_hour
                    # monthly average conditions between cooling peak and heating peak
                    if self.monthly_peak_hl[i] > 0 and ipf[i]:

                        # heating peak
                        # self.load = np.append(self.load, self.monthly_peak_hl[i]) JDS corrected 20200604

                        self.load = np.append(self.load, -self.monthly_peak_hl[i])
                        self.hour = np.append(
                            self.hour,
                            last_hour_heating_peak
                            + self.monthly_peak_hl_duration[i] / 2,
                        )

                        if last_avg_hour - peak_last_avg_hour < 0.0:
                            warnings.warn(warn_msg_neg_timestep)
                        peak_last_avg_hour = last_avg_hour
                    # rest of month
                    last_avg_hour = last_month_hour(i, self.years)
                    self.load = np.append(self.load, month_rate)
                    self.hour = np.append(self.hour, last_avg_hour)

                    if last_avg_hour - peak_last_avg_hour < 0.0:
                        warnings.warn(warn_msg_neg_timestep)
                    peak_last_avg_hour = last_avg_hour

                else:
                    last_avg_hour = last_month_hour(i, self.years)
                    self.load = np.append(self.load, month_rate)
                    self.hour = np.append(self.hour, last_avg_hour)

                if last_avg_hour - peak_last_avg_hour < 0.0:
                    warnings.warn(warn_msg_neg_timestep)
                peak_last_avg_hour = last_avg_hour

        #       Now fill array containing step function loads
        #       Note they are paired with the ending hour, so the ith load will start with the (i-1)th time

        n = self.hour.size
        # Note at this point the load and hour np arrays contain zeroes in indices zero and one, then continue from
        # there.
        for i in range(1, n):
            step_load = self.load[i] - self.load[i - 1]
            self.step_func_load = np.append(self.step_func_load, step_load)
        pass


def number_to_month(x):
    return ["NULL", "January", "February", "March", "April", "May", "June",
            "July", "August", "September", "October", "November", "December"][x]


def monthdays(month, year):
    leap_year = year % 4 == 0
    if month > 12:
        md = month % 12
    else:
        md = month
    if leap_year:
        num_days = [31, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    else:
        num_days = [31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    return num_days[md]


def first_month_hour(month, years):
    fmh = 1
    if month > 1:
        for i in range(1, month):
            if len(years) > 1:
                current_year = years[(month - 1) // 12]
            else:
                current_year = years[0]
            mi = i % 12
            fmh = fmh + 24 * monthdays(mi, current_year)
    return fmh


def last_month_hour(month, years):
    lmh = 0
    for i in range(1, month + 1):
        if len(years) > 1:
            current_year = years[(month - 1) // 12]
        else:
            current_year = years[0]
        lmh = lmh + monthdays(i, current_year) * 24
    if month == 1:
        lmh = 31 * 24
    return lmh



from ghedesigner.shape import point_polygon_check


def remove_cutout(coordinates, boundaries, remove_inside=True, keep_contour=True, on_edge_tolerance=0.01):

    if isinstance(boundaries[0][0], (int, float)):
        boundaries = [boundaries]

    new_coordinates = []
    inside = 1
    on_edge = 0
    for idx, coordinate in enumerate(coordinates):
        coordinate = coordinates[idx]
        boundary_results = []
        for boundary in boundaries:
            boundary_results.append(point_polygon_check(boundary, coordinate, on_edge_tolerance=on_edge_tolerance))
        if remove_inside:
            if (inside not in boundary_results) and not (on_edge in boundary_results and not keep_contour):
                new_coordinates.append(coordinate)
        else:
            if (inside in boundary_results) or (on_edge in boundary_results and keep_contour):
                new_coordinates.append(coordinate)

    return new_coordinates


def determine_largest_rectangle(property_boundary):
    x_max = float('-inf')
    y_max = float('-inf')
    x_min = float('inf')
    y_min = float('inf')
    for bf_outline in property_boundary:
        for x, y in bf_outline:
            if x > x_max:
                x_max = x
            if y > y_max:
                y_max = y
            if x < x_min:
                x_min = x
            if y < y_min:
                y_min = y

    rectangle = [[x_min, y_min], [x_max, y_min], [x_max, y_max], [x_min, y_max], [x_min, y_min]]

    return rectangle



from math import ceil, sqrt
from typing import Optional

from ghedesigner.borehole_heat_exchangers import GHEBorehole
from ghedesigner.enums import BHPipeType, TimestepType, FlowConfigType
from ghedesigner.gfunction import calc_g_func_for_multiple_lengths
from ghedesigner.ground_heat_exchangers import GHE
from ghedesigner.media import Grout, Pipe, Soil, GHEFluid
from ghedesigner.rowwise import field_optimization_fr, field_optimization_wp_space_fr, gen_shape
from ghedesigner.simulation import SimulationParameters
from ghedesigner.utilities import eskilson_log_times, borehole_spacing, check_bracket, sign


class Bisection1D:
    def __init__(
            self,
            coordinates_domain: list,
            field_descriptors: list,
            v_flow: float,
            borehole: GHEBorehole,
            bhe_type: BHPipeType,
            fluid: GHEFluid,
            pipe: Pipe,
            grout: Grout,
            soil: Soil,
            sim_params: SimulationParameters,
            hourly_extraction_ground_loads: list,
            method: TimestepType,
            flow_type: FlowConfigType.BOREHOLE,
            max_iter=15,
            disp=False,
            search=True,
            field_type="N/A",
            load_years=None,
    ):

        # Take the lowest part of the coordinates domain to be used for the
        # initial setup
        if load_years is None:
            load_years = [2019]
        self.load_years = load_years
        self.searchTracker = []
        coordinates = coordinates_domain[0]
        current_field = field_descriptors[0]
        self.field_type = field_type
        # Flow rate tracking
        self.V_flow = v_flow
        self.flow_type = flow_type
        v_flow_system, m_flow_borehole = self.retrieve_flow(coordinates, fluid.rho)
        self.method = method

        self.log_time = eskilson_log_times()
        self.bhe_type = bhe_type
        self.sim_params = sim_params
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.coordinates_domain = coordinates_domain
        self.fieldDescriptors = field_descriptors
        self.max_iter = max_iter
        self.disp = disp

        b = borehole_spacing(borehole, coordinates)

        # Calculate a g-function for uniform inlet fluid temperature with
        # 8 unequal segments using the equivalent solver
        g_function = calc_g_func_for_multiple_lengths(
            b,
            [borehole.H],
            borehole.r_b,
            borehole.D,
            m_flow_borehole,
            self.bhe_type,
            self.log_time,
            coordinates,
            fluid,
            pipe,
            grout,
            soil,
        )

        # Initialize the GHE object
        self.ghe = GHE(
            v_flow_system,
            b,
            bhe_type,
            fluid,
            borehole,
            pipe,
            grout,
            soil,
            g_function,
            sim_params,
            hourly_extraction_ground_loads,
            field_specifier=current_field,
            field_type=field_type,
            load_years=load_years,
        )

        self.calculated_temperatures = {}

        if search:
            self.selection_key, self.selected_coordinates = self.search()

    def retrieve_flow(self, coordinates, rho):
        if self.flow_type == FlowConfigType.BOREHOLE:
            v_flow_system = self.V_flow * len(coordinates)
            # Total fluid mass flow rate per borehole (kg/s)
            m_flow_borehole = self.V_flow / 1000.0 * rho
        elif self.flow_type == FlowConfigType.SYSTEM:
            v_flow_system = self.V_flow
            v_flow_borehole = self.V_flow / len(coordinates)
            m_flow_borehole = v_flow_borehole / 1000.0 * rho
        else:
            raise ValueError("The flow argument should be either `borehole`" "or `system`.")
        return v_flow_system, m_flow_borehole

    def initialize_ghe(self, coordinates, h, field_specifier="N/A"):
        v_flow_system, m_flow_borehole = self.retrieve_flow(coordinates, self.ghe.bhe.fluid.rho)

        self.ghe.bhe.b.H = h
        borehole = self.ghe.bhe.b
        fluid = self.ghe.bhe.fluid
        pipe = self.ghe.bhe.pipe
        grout = self.ghe.bhe.grout
        soil = self.ghe.bhe.soil

        b = borehole_spacing(borehole, coordinates)

        # Calculate a g-function for uniform inlet fluid temperature with
        # 8 unequal segments using the equivalent solver
        g_function = calc_g_func_for_multiple_lengths(
            b,
            [borehole.H],
            borehole.r_b,
            borehole.D,
            m_flow_borehole,
            self.bhe_type,
            self.log_time,
            coordinates,
            fluid,
            pipe,
            grout,
            soil,
        )

        # Initialize the GHE object
        self.ghe = GHE(
            v_flow_system,
            b,
            self.bhe_type,
            fluid,
            borehole,
            pipe,
            grout,
            soil,
            g_function,
            self.sim_params,
            self.hourly_extraction_ground_loads,
            field_type=self.field_type,
            field_specifier=field_specifier,
            load_years=self.load_years,
        )

    def calculate_excess(self, coordinates, h, field_specifier="N/A"):
        self.initialize_ghe(coordinates, h, field_specifier=field_specifier)
        # Simulate after computing just one g-function
        max_hp_eft, min_hp_eft = self.ghe.simulate(method=self.method)
        t_excess = self.ghe.cost(max_hp_eft, min_hp_eft)
        self.searchTracker.append([field_specifier, t_excess, max_hp_eft, min_hp_eft])

        return t_excess

    def search(self):

        x_l_idx = 0

        # find upper bound that respects max_boreholes
        if self.sim_params.max_boreholes is not None:
            num_coordinates_in_each = [len(x) for x in self.coordinates_domain]
            x_r_idx = [idx for idx, x in enumerate(num_coordinates_in_each) if x < self.sim_params.max_boreholes][-1]
        else:
            x_r_idx = len(self.coordinates_domain) - 1

        if self.disp:
            print("Do some initial checks before searching.")
        # Get the lowest possible excess temperature from minimum height at the
        # smallest location in the domain
        t_0_lower = self.calculate_excess(
            self.coordinates_domain[x_l_idx],
            self.sim_params.min_height,
            field_specifier=self.fieldDescriptors[x_l_idx],
        )
        t_0_upper = self.calculate_excess(
            self.coordinates_domain[x_l_idx],
            self.sim_params.max_height,
            field_specifier=self.fieldDescriptors[x_l_idx],
        )
        t_m1 = self.calculate_excess(
            self.coordinates_domain[x_r_idx],
            self.sim_params.max_height,
            field_specifier=self.fieldDescriptors[x_r_idx],
        )

        self.calculated_temperatures[x_l_idx] = t_0_upper
        self.calculated_temperatures[x_r_idx] = t_m1

        if check_bracket(sign(t_0_lower), sign(t_0_upper)):
            if self.disp:
                print("Size between min and max of lower bound in domain.")
            self.initialize_ghe(self.coordinates_domain[x_l_idx], self.sim_params.max_height)
            return x_l_idx, self.coordinates_domain[x_l_idx]
        elif check_bracket(sign(t_0_upper), sign(t_m1)):
            if self.disp:
                print("Perform the integer bisection search routine.")
            pass
        else:
            # solution doesn't lie within bounds
            if t_0_lower < 0.0:
                condition_msg = "The optimal design requires fewer or shorter boreholes \n" \
                                "than what is possible based on the current design parameters."
                print(condition_msg)
                if self.sim_params.continue_if_design_unmet:
                    print("Smallest available configuration selected.")
                    selection_key = x_l_idx
                    self.initialize_ghe(self.coordinates_domain[selection_key], self.sim_params.min_height,
                                        self.fieldDescriptors[selection_key])
                    return selection_key, self.coordinates_domain[selection_key]
                else:
                    raise ValueError("Search failed.")
            elif t_m1 > 0.0:
                condition_msg = "The optimal design requires more or deeper boreholes \n" \
                                "than what is possible based on the current design parameters. \n" \
                                "Consider increasing the available land area, \n" \
                                "increasing the maximum borehole depth, \n" \
                                "or decreasing the maximum borehole spacing."
                print(condition_msg)
                if self.sim_params.continue_if_design_unmet:
                    print("Largest available configuration selected.")
                    selection_key = x_r_idx
                    self.initialize_ghe(self.coordinates_domain[selection_key], self.sim_params.max_height,
                                        self.fieldDescriptors[selection_key])
                    return selection_key, self.coordinates_domain[selection_key]
                else:
                    raise ValueError("Search failed.")
            else:
                # if we've gotten here, everything should be good for the bisection search.
                # can add catches for other cases here if they pop up.
                pass
        if self.disp:
            print("Beginning bisection search...")

        x_l_sign = sign(t_0_upper)

        i = 0

        while i < self.max_iter:
            c_idx = ceil((x_l_idx + x_r_idx) / 2)
            # if the solution is no longer making progress break the while
            if c_idx == x_l_idx or c_idx == x_r_idx:
                break

            c_t_excess = self.calculate_excess(
                self.coordinates_domain[c_idx],
                self.sim_params.max_height,
                field_specifier=self.fieldDescriptors[c_idx],
            )

            self.calculated_temperatures[c_idx] = c_t_excess
            c_sign = sign(c_t_excess)

            if c_sign == x_l_sign:
                x_l_idx = c_idx
            else:
                x_r_idx = c_idx

            i += 1

        coordinates = self.coordinates_domain[i]

        self.calculate_excess(coordinates, self.sim_params.max_height, self.fieldDescriptors[i])
        # Make sure the field being returned pertains to the index which is the
        # closest to 0 but also negative (the maximum of all 0 or negative
        # excess temperatures)
        keys = list(self.calculated_temperatures.keys())
        values = list(self.calculated_temperatures.values())

        # theoretically, the biggest negative value should be the field that is just undersized
        negative_excess_values = [v for v in values if v <= 0.0]
        excess_of_interest = max(negative_excess_values)

        # but some conditions don't yield this result
        # adding a check here to ensure we pick the smallest field with
        # negative excess temperature
        num_bh = [len(self.coordinates_domain[x]) for x in keys]
        sorted_num_bh, sorted_values = (list(t) for t in zip(*sorted(zip(num_bh, values))))
        for _, val in zip(sorted_num_bh, sorted_values):
            if val < 0:
                if excess_of_interest != val:
                    print('Loads resulted in odd behavior requiring the selected field configuration \n'
                          'to be reset to the smallest field with negative excess temperature. \n'
                          'Please forward the inputs to the developers for investigation.')
                excess_of_interest = val
                break

        idx = values.index(excess_of_interest)
        selection_key = keys[idx]
        self.initialize_ghe(self.coordinates_domain[selection_key], self.sim_params.max_height,
                            self.fieldDescriptors[selection_key])
        return selection_key, self.coordinates_domain[selection_key]


# This is the search algorithm used for finding row-wise fields
class RowWiseModifiedBisectionSearch:
    def __init__(
            self,
            v_flow: float,
            borehole: GHEBorehole,
            bhe_type: BHPipeType,
            fluid: GHEFluid,
            pipe: Pipe,
            grout: Grout,
            soil: Soil,
            sim_params: SimulationParameters,
            hourly_extraction_ground_loads: list,
            geometric_constraints,
            method: TimestepType,
            flow_type: FlowConfigType.BOREHOLE,
            max_iter: int = 10,
            disp: bool = False,
            search: bool = True,
            advanced_tracking: bool = True,
            field_type: str = "rowwise",
            load_years=None,
    ):

        # Take the lowest part of the coordinates domain to be used for the
        # initial setup
        if load_years is None:
            load_years = [2019]
        self.load_years = load_years
        self.fluid = fluid
        self.pipe = pipe
        self.grout = grout
        self.soil = soil
        self.borehole = borehole
        self.geometricConstraints = geometric_constraints
        self.searchTracker = []
        self.fieldType = field_type
        # Flow rate tracking
        self.V_flow = v_flow
        self.flow_type = flow_type
        self.method = method
        self.log_time = eskilson_log_times()
        self.bhe_type = bhe_type
        self.sim_params = sim_params
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.max_iter = max_iter
        self.disp = disp
        self.ghe: Optional[GHE] = None
        self.calculated_temperatures = {}
        if advanced_tracking:
            self.advanced_tracking = [["TargetSpacing", "Field Specifier", "nbh", "ExcessTemperature"]]
            self.checkedFields = []
        if search:
            self.selected_coordinates, self.selected_specifier = self.search()
            self.initialize_ghe(self.selected_coordinates, self.sim_params.max_height,
                                field_specifier=self.selected_specifier)

    def retrieve_flow(self, coordinates, rho):
        if self.flow_type == FlowConfigType.BOREHOLE:
            v_flow_system = self.V_flow * len(coordinates)
            # Total fluid mass flow rate per borehole (kg/s)
            m_flow_borehole = self.V_flow / 1000.0 * rho
        elif self.flow_type == FlowConfigType.SYSTEM:
            v_flow_system = self.V_flow
            v_flow_borehole = self.V_flow / len(coordinates)
            m_flow_borehole = v_flow_borehole / 1000.0 * rho
        else:
            raise ValueError("The flow argument should be either `borehole`" "or `system`.")
        return v_flow_system, m_flow_borehole

    def initialize_ghe(self, coordinates, h, field_specifier="N/A"):
        v_flow_system, m_flow_borehole = self.retrieve_flow(coordinates, self.fluid.rho)

        self.borehole.H = h
        borehole = self.borehole
        fluid = self.fluid
        pipe = self.pipe
        grout = self.grout
        soil = self.soil

        b = borehole_spacing(borehole, coordinates)

        # Calculate a g-function for uniform inlet fluid temperature with
        # 8 unequal segments using the equivalent solver
        g_function = calc_g_func_for_multiple_lengths(
            b,
            [borehole.H],
            borehole.r_b,
            borehole.D,
            m_flow_borehole,
            self.bhe_type,
            self.log_time,
            coordinates,
            fluid,
            pipe,
            grout,
            soil,
        )

        # Initialize the GHE object
        self.ghe = GHE(
            v_flow_system,
            b,
            self.bhe_type,
            fluid,
            borehole,
            pipe,
            grout,
            soil,
            g_function,
            self.sim_params,
            self.hourly_extraction_ground_loads,
            field_type=self.fieldType,
            field_specifier=field_specifier,
            load_years=self.load_years,
        )

    def calculate_excess(self, coordinates, h, field_specifier="N/A"):
        self.initialize_ghe(coordinates, h, field_specifier=field_specifier)
        # Simulate after computing just one g-function
        max_hp_eft, min_hp_eft = self.ghe.simulate(method=self.method)
        t_excess = self.ghe.cost(max_hp_eft, min_hp_eft)
        self.searchTracker.append([field_specifier, t_excess, max_hp_eft, min_hp_eft])

        return t_excess

    def search(self):

        spacing_start = self.geometricConstraints.min_spacing
        spacing_stop = self.geometricConstraints.max_spacing
        spacing_step = self.geometricConstraints.spacing_step
        rotate_step = self.geometricConstraints.rotate_step
        prop_bound, ng_zones = gen_shape(self.geometricConstraints.property_boundary,
                                         self.geometricConstraints.no_go_boundaries)
        rotate_start = self.geometricConstraints.min_rotation
        rotate_stop = self.geometricConstraints.max_rotation
        perimeter_spacing_ratio = self.geometricConstraints.perimeter_spacing_ratio

        if perimeter_spacing_ratio is None:
            use_perimeter = False
        else:
            use_perimeter = True

        selected_coordinates = None
        selected_specifier = None
        selected_temp_excess = None
        selected_spacing = None

        # Check The Upper and Lower Bounds

        # Generate Fields
        if use_perimeter:
            upper_field, upper_field_specifier = field_optimization_wp_space_fr(
                perimeter_spacing_ratio,
                spacing_start,
                rotate_step,
                prop_bound,
                ng_zones=ng_zones,
                rotate_start=rotate_start,
                rotate_stop=rotate_stop,
            )
            lower_field, lower_field_specifier = field_optimization_wp_space_fr(
                perimeter_spacing_ratio,
                spacing_stop,
                rotate_step,
                prop_bound,
                ng_zones=ng_zones,
                rotate_start=rotate_start,
                rotate_stop=rotate_stop,
            )
        else:
            upper_field, upper_field_specifier = field_optimization_fr(
                spacing_start,
                rotate_step,
                prop_bound,
                ng_zones=ng_zones,
                rotate_start=rotate_start,
                rotate_stop=rotate_stop,
            )
            lower_field, lower_field_specifier = field_optimization_fr(
                spacing_stop,
                rotate_step,
                prop_bound,
                ng_zones=ng_zones,
                rotate_start=rotate_start,
                rotate_stop=rotate_stop,
            )

        # Get Excess Temperatures
        t_upper = self.calculate_excess(upper_field, self.sim_params.max_height, field_specifier=upper_field_specifier)
        t_lower = self.calculate_excess(lower_field, self.sim_params.max_height, field_specifier=lower_field_specifier)

        if self.advanced_tracking:
            self.advanced_tracking.append([spacing_start, upper_field_specifier, len(upper_field), t_upper])
            self.advanced_tracking.append([spacing_stop, lower_field_specifier, len(lower_field), t_lower])
            self.checkedFields.append(upper_field)
            self.checkedFields.append(lower_field)

        # If the excess temperature is >0 utilizing the largest field and largest depth, then notify the user that
        # the given constraints cannot find a satisfactory field.
        if t_upper > 0.0 and t_lower > 0.0:
            condition_msg = "The optimal design requires more or deeper boreholes \n" \
                            "than what is possible based on the current design parameters. \n" \
                            "Consider increasing the available land area, \n" \
                            "increasing the maximum borehole depth, \n" \
                            "or decreasing the maximum borehole spacing."
            print(condition_msg)
            if self.sim_params.continue_if_design_unmet:
                print("Largest available configuration selected.")
                return upper_field, upper_field_specifier
            else:
                raise ValueError("Search failed.")

        # If the excess temperature is > 0 when utilizing the largest field and depth but < 0 when using the largest
        # depth and smallest field, then fields should be searched between the two target depths.
        elif t_upper < 0.0 < t_lower:
            # This search currently works by doing a slightly modified bisection search where the "steps" are the set
            # by the "spacing_step" variable. The steps are used to check fields on either side of the field found by
            # doing a normal bisection search. These extra fields are meant to help prevent falling into local minima
            # (although this will still happen sometimes).
            i = 0
            spacing_high = spacing_start
            spacing_low = spacing_stop
            low_e = t_upper
            high_e = t_lower
            spacing_m = (spacing_stop + spacing_start) * 0.5
            while i < self.max_iter:
                print("Bisection Search Iteration: ", i)
                # Getting Three Middle Field
                if use_perimeter:
                    f1, f1_specifier = field_optimization_wp_space_fr(
                        perimeter_spacing_ratio,
                        spacing_m,
                        rotate_step,
                        prop_bound,
                        ng_zones=ng_zones,
                        rotate_start=rotate_start,
                        rotate_stop=rotate_stop,
                    )
                else:
                    f1, f1_specifier = field_optimization_fr(
                        spacing_m,
                        rotate_step,
                        prop_bound,
                        ng_zones=ng_zones,
                        rotate_start=rotate_start,
                        rotate_stop=rotate_stop,
                    )

                # Getting the three field's excess temperature
                t_e1 = self.calculate_excess(
                    f1, self.sim_params.max_height, field_specifier=f1_specifier
                )

                if self.advanced_tracking:
                    self.advanced_tracking.append([spacing_m, f1_specifier, len(f1), t_e1])
                    self.checkedFields.append(f1)
                if t_e1 <= 0.0:
                    spacing_high = spacing_m
                    high_e = t_e1
                    selected_specifier = f1_specifier
                else:
                    spacing_low = spacing_m
                    low_e = t_e1

                spacing_m = (spacing_low + spacing_high) * 0.5
                if abs(low_e - high_e) < 1E-10:  # Error tolerance
                    break

                i += 1

            # Now Check fields that have a higher target spacing to double-check that none of them would work:
            spacing_l = spacing_step + spacing_high
            target_spacings = []
            current_spacing = spacing_high

            # TODO: this was an argument, but was never used.
            exhaustive_fields_to_check = 10
            spacing_change = (spacing_l - current_spacing) / exhaustive_fields_to_check
            while current_spacing <= spacing_l:
                target_spacings.append(current_spacing)
                current_spacing += spacing_change
            best_field = None
            best_drilling = float("inf")
            best_excess = None
            best_spacing = None
            for ts in target_spacings:
                if use_perimeter:
                    field, f_s = field_optimization_wp_space_fr(
                        perimeter_spacing_ratio,
                        ts,
                        rotate_step,
                        prop_bound,
                        ng_zones=ng_zones,
                        rotate_start=rotate_start,
                        rotate_stop=rotate_stop,
                    )
                else:
                    field, f_s = field_optimization_fr(
                        ts,
                        rotate_step,
                        prop_bound,
                        ng_zones=ng_zones,
                        rotate_start=rotate_start,
                        rotate_stop=rotate_stop,
                    )

                t_e = self.calculate_excess(field, self.sim_params.max_height, field_specifier=f_s)

                if self.advanced_tracking:
                    self.advanced_tracking.append([ts, f_s, len(field), t_e])
                    self.checkedFields.append(field)

                self.initialize_ghe(field, self.sim_params.max_height, field_specifier=f_s)
                self.ghe.compute_g_functions()
                self.ghe.size(method=TimestepType.HYBRID)
                total_drilling = self.ghe.bhe.b.H * len(field)

                if best_field is None:
                    best_field = field
                    best_drilling = total_drilling
                    best_excess = t_e
                    best_spacing = ts
                else:
                    if t_e <= 0.0 and total_drilling < best_drilling:
                        best_drilling = total_drilling
                        best_field = field
                        best_excess = t_e
                        best_spacing = ts
            selected_coordinates = best_field
            selected_temp_excess = best_excess
            selected_spacing = best_spacing

        # If the excess temperature is < 0 when utilizing the largest depth and the smallest field, it is most likely
        # in the user's best interest to return a field smaller than the smallest one. This is done by removing
        # boreholes from the field.
        elif t_lower < 0.0 and t_upper < 0.0:

            original_coordinates = lower_field

            # Function For Sorting Boreholes Based on Proximity to a Point
            def point_sort(target_point, other_points, method="ascending"):
                def dist(o_p):
                    return sqrt(
                        (target_point[0] - o_p[0]) * (target_point[0] - o_p[0])
                        + (target_point[1] - o_p[1]) * (target_point[1] - o_p[1]))

                distances = map(dist, other_points)
                if method == "ascending":
                    return [x for _, x in sorted(zip(distances, other_points))]
                elif method == "descending":
                    return [x for _, x in sorted(zip(distances, other_points), reverse=True)]

            # TODO: b_r_removal_method was an argument but it was never used
            # if b_r_removal_method == "CloseToCorner":
            starting_field = point_sort(original_coordinates[0], lower_field, method="descending")
            # elif b_r_removal_method == "CloseToPoint":
            #     starting_field = point_sort([0.0, 0.0], lower_field, method="descending")
            # elif b_r_removal_method == "FarFromPoint":
            #     starting_field = point_sort([0.0, 0.0], lower_field, method="ascending")
            # elif b_r_removal_method == "RowRemoval":
            #     starting_field = lower_field
            # else:
            #     msg = b_r_removal_method + " is not a valid method for removing boreholes."
            #     msg += "The valid methods are: CloseToCorner, CloseToPoint, FarFromPoint, and RowRemoval."
            #     raise ValueError(msg)

            # Check if a 1X1 field is satisfactory
            t_e_single = self.calculate_excess([[0, 0]], self.sim_params.max_height, field_specifier="1X1")

            if self.advanced_tracking:
                self.advanced_tracking.append(["N/A", "1X1", 1, t_e_single])
                self.checkedFields.append([[0, 0]])
            if t_e_single <= 0:
                selected_temp_excess = t_e_single
                selected_specifier = "1X1"
                selected_coordinates = starting_field[len(starting_field) - 1:]
                selected_spacing = spacing_stop
            else:
                # Perform a bisection search between nbh values to find the smallest satisfactory field
                nbh_max = len(starting_field)
                nbh_min = 1
                nbh_start = nbh_max
                # continueLoop = True
                # highT_e = T_lower
                selected_specifier = lower_field_specifier
                i = 0
                while i < self.max_iter:
                    nbh = (nbh_max + nbh_min) // 2
                    current_field = starting_field[nbh_start - nbh:]
                    f_s = lower_field_specifier + f"_BR{nbh_start - nbh}"
                    t_e = self.calculate_excess(
                        current_field, self.sim_params.max_height, field_specifier=f_s
                    )
                    if self.advanced_tracking:
                        self.advanced_tracking.append(
                            [spacing_stop, lower_field_specifier + "_" + str(nbh), nbh, t_e]
                        )
                        self.checkedFields.append(current_field)
                    if t_e <= 0.0:
                        # highT_e = T_e
                        nbh_max = nbh
                        selected_coordinates = current_field
                        selected_specifier = f_s
                        selected_temp_excess = t_e
                        selected_spacing = spacing_stop
                    else:
                        nbh_min = nbh
                    if (nbh_max - nbh_min) <= 1:
                        break
                    i += 1
        # If none of the options above have been true, then there is most likely an issue with the excess temperature
        # calculation.
        else:
            msg = (
                "There seems to be an issue calculating excess temperatures. Check that you have the correct \n"
                "package version. If this is a recurring issue, please contact the current package management for \n"
                "assistance."
            )
            raise ValueError(msg)
        if self.advanced_tracking:
            self.advanced_tracking.append(
                [
                    selected_spacing,
                    selected_specifier,
                    len(selected_coordinates),
                    selected_temp_excess,
                ]
            )
            self.checkedFields.append(selected_coordinates)
        return selected_coordinates, selected_specifier


class Bisection2D(Bisection1D):
    def __init__(
            self,
            coordinates_domain_nested: list,
            field_descriptors: list,
            v_flow: float,
            borehole: GHEBorehole,
            bhe_type,
            fluid: GHEFluid,
            pipe: Pipe,
            grout: Grout,
            soil: Soil,
            sim_params: SimulationParameters,
            hourly_extraction_ground_loads: list,
            method: TimestepType,
            flow_type: FlowConfigType.BOREHOLE,
            max_iter=15,
            disp=False,
            field_type="N/A",
            load_years=None,
    ):
        if load_years is None:
            load_years = [2019]
        if disp:
            print("Note: This routine requires a nested bisection search.")
        self.load_years = load_years
        # Get a coordinates domain for initialization
        coordinates_domain = coordinates_domain_nested[0]
        super().__init__(
            coordinates_domain,
            field_descriptors[0],
            v_flow,
            borehole,
            bhe_type,
            fluid,
            pipe,
            grout,
            soil,
            sim_params,
            hourly_extraction_ground_loads,
            method=method,
            flow_type=flow_type,
            max_iter=max_iter,
            disp=disp,
            search=False,
            field_type=field_type,
            load_years=load_years,
        )

        self.coordinates_domain_nested = []
        self.calculated_temperatures_nested = []
        # Tack on one borehole at the beginning to provide a high excess temperature
        outer_domain = [coordinates_domain_nested[0][0]]
        for cdn in coordinates_domain_nested:
            outer_domain.append(cdn[-1])

        self.coordinates_domain = outer_domain

        selection_key, _ = self.search()

        self.calculated_temperatures_nested.append(self.calculated_temperatures)

        # We tacked on one borehole to the beginning, so we need to subtract 1
        # on the index
        inner_domain = coordinates_domain_nested[selection_key - 1]
        self.coordinates_domain = inner_domain
        self.fieldDescriptors = field_descriptors[selection_key - 1]

        # Reset calculated temperatures
        self.calculated_temperatures = {}

        self.selection_key, self.selected_coordinates = self.search()


class BisectionZD(Bisection1D):
    def __init__(
            self,
            coordinates_domain_nested: list,
            field_descriptors: list,
            v_flow: float,
            borehole: GHEBorehole,
            bhe_type,
            fluid: GHEFluid,
            pipe: Pipe,
            grout: Grout,
            soil: Soil,
            sim_params: SimulationParameters,
            hourly_extraction_ground_loads: list,
            method: TimestepType,
            flow_type: FlowConfigType.BOREHOLE,
            max_iter=15,
            disp=False,
            field_type="N/A",
            load_years=None,
    ):
        if load_years is None:
            load_years = [2019]
        if disp:
            print("Note: This design routine currently requires several bisection searches.")

        # Get a coordinates domain for initialization
        coordinates_domain = coordinates_domain_nested[0]
        super().__init__(
            coordinates_domain,
            field_descriptors[0],
            v_flow,
            borehole,
            bhe_type,
            fluid,
            pipe,
            grout,
            soil,
            sim_params,
            hourly_extraction_ground_loads,
            method=method,
            flow_type=flow_type,
            max_iter=max_iter,
            disp=disp,
            search=False,
            field_type=field_type,
            load_years=load_years,
        )

        self.coordinates_domain_nested = coordinates_domain_nested
        self.nested_fieldDescriptors = field_descriptors
        self.calculated_temperatures_nested = {}
        # Tack on one borehole at the beginning to provide a high excess
        # temperature
        outer_domain = [coordinates_domain_nested[0][0]]
        outer_descriptors = [field_descriptors[0][0]]
        for cdn, fd in zip(coordinates_domain_nested, field_descriptors):
            outer_domain.append(cdn[-1])
            outer_descriptors.append(fd[-1])

        self.coordinates_domain = outer_domain
        self.fieldDescriptors = outer_descriptors

        self.selection_key_outer, _ = self.search()
        if self.selection_key_outer > 0:
            self.selection_key_outer -= 1
        self.calculated_heights = {}

        self.selection_key, self.selected_coordinates = self.search_successive()

    def search_successive(self, max_iter=None):
        if max_iter is None:
            max_iter = self.selection_key_outer + 7

        i = self.selection_key_outer

        old_height = 99999

        while i < len(self.coordinates_domain_nested) and i < max_iter:

            self.coordinates_domain = self.coordinates_domain_nested[i]
            self.fieldDescriptors = self.nested_fieldDescriptors[i]
            self.calculated_temperatures = {}
            try:
                selection_key, selected_coordinates = self.search()
            except ValueError:
                break
            self.calculated_temperatures_nested[i] = self.calculated_temperatures

            self.ghe.compute_g_functions()
            self.ghe.size(method=TimestepType.HYBRID)

            nbh = len(selected_coordinates)
            total_drilling = nbh * self.ghe.bhe.b.H
            self.calculated_heights[i] = total_drilling

            if old_height < total_drilling:
                break
            else:
                old_height = total_drilling

            i += 1

        keys = list(self.calculated_heights.keys())
        values = list(self.calculated_heights.values())

        minimum_total_drilling = min(values)
        idx = values.index(minimum_total_drilling)
        selection_key_outer = keys[idx]
        self.calculated_temperatures = self.calculated_temperatures_nested[selection_key_outer]

        keys = list(self.calculated_temperatures.keys())
        values = list(self.calculated_temperatures.values())

        negative_excess_values = [v for v in values if v <= 0.0]

        excess_of_interest = max(negative_excess_values)
        idx = values.index(excess_of_interest)
        selection_key = keys[idx]
        selected_coordinates = self.coordinates_domain_nested[selection_key_outer][
            selection_key
        ]

        self.initialize_ghe(
            selected_coordinates,
            self.sim_params.max_height,
            field_specifier=self.nested_fieldDescriptors[selection_key_outer][
                selection_key
            ],
        )
        self.ghe.compute_g_functions()
        self.ghe.size(method=TimestepType.HYBRID)

        return selection_key, selected_coordinates



from pathlib import Path

from setuptools import setup

from ghedesigner import VERSION

readme_file = Path(__file__).parent.resolve() / 'README.md'
readme_contents = readme_file.read_text(encoding='utf8')

short_description = """A ground heat exchanger design tool with the capability
to select and size flexibly configured borehole fields that are customized
for specific building and property constraints."""

setup(
    name='GHEDesigner',
    install_requires=[
        'click~=8.1',
        'jsonschema~=4.19',
        'numpy~=1.24',
        'pygfunction~=2.2',
        'scipy~=1.10'
    ],
    url='https://github.com/BETSRG/GHEDesigner',
    description=short_description,
    license='BSD-3',
    long_description=readme_contents,
    long_description_content_type='text/markdown',
    version=VERSION,
    packages=['ghedesigner'],
    include_package_data=True,
    package_data={'ghedesigner': ['schemas/*.json']},
    author='Jeffrey D. Spitler',
    author_email='spitler@okstate.edu',
    entry_points={
        'console_scripts': ['ghedesigner=ghedesigner.manager:run_manager_from_cli']
    },
    python_requires='>=3.8',
    classifiers=[
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11'
    ]
)
